#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Interactive script for analysis of Tardieu trials.
matplotlib + Qt5

TODO:

    accmeter chs should be selectable from gui (if used)
    add accmeter ch label
    buttons become active if first data load fails
    filter behaves weirdly when hp is set at 0 Hz
    filter crash when lp is set at 1000 Hz
    no narrow view

    params to config
    EMG scale box is not updated when scaling w/ toolbar tool
    add statusbar?


@author: Jussi (jnu@iki.fi)
"""

from __future__ import print_function
from __future__ import division
from builtins import zip
from builtins import str
from builtins import object
from collections import OrderedDict
import matplotlib
from matplotlib.figure import Figure
import matplotlib.gridspec as gridspec
import logging
import sys
import numpy as np
import copy
from pkg_resources import resource_filename
from PyQt5 import QtGui, QtWidgets, uic, QtCore
from matplotlib.backends.backend_qt5agg import (FigureCanvasQTAgg,
                                                NavigationToolbar2QT)
from matplotlib.backends.backend_pdf import PdfPages
from matplotlib.backends.backend_agg import FigureCanvasAgg as FigureCanvas

from .. import (EMG, nexus, cfg, read_data, trial, eclipse, models,
                Trial, Plotter, layouts, utils, GaitDataError,
                register_gui_exception_handler)
from ..numutils import segment_angles, rms
from .qt_dialogs import qt_message_dialog, qt_yesno_dialog


matplotlib.style.use(cfg.plot.mpl_style)

logger = logging.getLogger(__name__)


def read_nexus_starting_angle():
    """Read the Nexus defined starting angle"""
    subjname = nexus.get_subjectnames()
    vicon = nexus.viconnexus()
    asp = vicon.GetSubjectParam(subjname, 'AnkleStartPos')
    return asp[0] if asp[1] else None


class LoadDialog(QtWidgets.QDialog):
    """ Dialog for loading data """

    def __init__(self):

        super(self.__class__, self).__init__()
        uifile = resource_filename('gaitutils', 'gui/tardieu_load_dialog.ui')
        uic.loadUi(uifile, self)
        try:
            ang0_nexus = read_nexus_starting_angle()
        except GaitDataError:
            ang0_nexus = None
        self.spNormAngle.setValue(ang0_nexus if ang0_nexus else 90)


class EMGFilterDialog(QtWidgets.QDialog):
    """ Dialog for setting the EMG filter """

    def __init__(self, emg_passband):

        super(self.__class__, self).__init__()
        uifile = resource_filename('gaitutils', 'gui/tardieu_filter_dialog.ui')
        uic.loadUi(uifile, self)
        self.spEMGLow.setValue(emg_passband[0])
        self.spEMGHigh.setValue(emg_passband[1])


class HelpDialog(QtWidgets.QDialog):
    """ Dialog for help"""

    def __init__(self):

        super(self.__class__, self).__init__()
        uifile = resource_filename('gaitutils', 'gui/tardieu_help_dialog.ui')
        uic.loadUi(uifile, self)


class SimpleToolbar(NavigationToolbar2QT):
    """ Simplified mpl navigation toolbar with some items removed """

    toolitems = [t for t in NavigationToolbar2QT.toolitems if
                 t[0] in ('Pan', 'Zoom')]


class TardieuWindow(QtWidgets.QMainWindow):
    """ Main Qt window with controls. The mpl figure containing the actual data
    is created by a separate class and embedded into this window. """
    #FIXME: all str casts are probably unnecessary

    def __init__(self, parent=None):

        super(TardieuWindow, self).__init__(parent)

        uifile = resource_filename('gaitutils', 'gui/tardieu.ui')
        uic.loadUi(uifile, self)

        self._tardieu_plot = TardieuPlot()
        # set the internal callbacks to point to our methods
        self._tardieu_plot._update_marker_status = self._update_marker_status
        self._tardieu_plot._update_status = self._update_status
        self.canvas = FigureCanvasQTAgg(self._tardieu_plot.fig)
        self.canvas.setSizePolicy(QtWidgets.QSizePolicy.Expanding,
                                  QtWidgets.QSizePolicy.Expanding)
        self.canvas.setParent(self)  # ?
        self.canvas.setFocusPolicy(QtCore.Qt.ClickFocus)
        self.canvas.show()

        self.btnClearMarkers.clicked.connect(self._clear_markers)
        self.btnSaveFig.clicked.connect(self._save_plot)
        self.btnZoomFast.clicked.connect(self._xzoom_to_fast)
        self.btnZoomReset.clicked.connect(self._xzoom_reset)

        self.spEMGYScale.valueChanged.connect(self._rescale_emg)
        self.btnSetEMGFilter.clicked.connect(self._emg_filter_dialog)

        self.emg_passband = cfg.emg.passband
        self.lblEMGPassband.setText('%.1f Hz - %.1f Hz' %
                                    (self.emg_passband[0],
                                     self.emg_passband[1]))

        # keep some controls disabled until data is loaded
        self._set_data_controls(False)
        self.btnQuit.clicked.connect(self.close)

        # no focus on buttons - need to keep focus on canvas for mpl events
        for w in self.findChildren(QtWidgets.QWidget):
            wname = str(w.objectName())
            if wname[:3] in ['btn']:
                w.setFocusPolicy(QtCore.Qt.NoFocus)

        self.actionQuit.triggered.connect(self.close)
        self.actionHelp.triggered.connect(self._help_dialog)
        self.actionOpen.triggered.connect(self._load_dialog_nexus)
        self.actionOpenC3D.triggered.connect(self._load_dialog_c3d)

        """
        # add the narrow view button?
        """
        self.lblStatus.setText('No data loaded')
        self.lblMarkerStatus.setText('No markers')

        # self.setStyleSheet("background-color: white;");
        # add canvas into last column, span all rows
        self.mainGridLayout.addWidget(self.canvas, 1,
                                      self.mainGridLayout.columnCount(),
                                      self.mainGridLayout.rowCount()-1, 1)

        # create toolbar and add it into last column, 1st row
        self.toolbar = SimpleToolbar(self.canvas, self)
        self._tardieu_plot._toolbar = self.toolbar
        self.mainGridLayout.addWidget(self.toolbar, 0,
                                      self.mainGridLayout.columnCount()-1,
                                      1, 1)
        self.canvas.setFocus()
        self.canvas.draw()

    def _xzoom_to_fast(self):
        self._tardieu_plot._xzoom_to_fast()
        self.canvas.draw()

    def _xzoom_reset(self):
        self._tardieu_plot._xzoom_reset()
        self.canvas.draw()

    def _reset_emg_filter(self, f1, f2):
        """Re-set the EMG filter"""
        self.emg_passband = [f1, f2]
        self.lblEMGPassband.setText('%.1f Hz - %.1f Hz' % (f1, f2))
        self._tardieu_plot._reset_emg_filter(f1, f2)
        self.canvas.draw()

    def _rescale_emg(self):
        """Callback for EMG rescaling"""
        yscale = self.spEMGYScale.value()
        self._tardieu_plot._rescale_emg(yscale)
        self.canvas.draw()
        self.canvas.setFocus()

    def _set_data_controls(self, enabled):
        """Show data related controls as (non)-responsive according to
        enabled (bool)"""
        for w in [self.btnSaveFig, self.spEMGYScale, self.btnClearMarkers,
                  self.btnSetEMGFilter, self.btnZoomFast, self.btnZoomReset]:
            w.setEnabled(enabled)

    def _nonresp(self):
        """Show whole window as non-responsive"""
        for w in self.findChildren(QtWidgets.QWidget):
            wname = str(w.objectName())
            if wname[:2] in ['bt', 'me', 'sp']:  # catch buttons, menus, spins
                w.setEnabled(False)
        # update display immediately in case thread gets blocked
        QtWidgets.QApplication.processEvents()
        QtWidgets.QApplication.setOverrideCursor(QtCore.Qt.WaitCursor)

    def _resp(self):
        """Show whole window as responsive"""
        for w in self.findChildren(QtWidgets.QWidget):
            wname = str(w.objectName())
            if wname[:2] in ['bt', 'me', 'sp']:
                w.setEnabled(True)
        QtWidgets.QApplication.restoreOverrideCursor()

    def _help_dialog(self):
        dlg = HelpDialog()
        dlg.exec_()

    def _load_dialog_nexus(self):
        """Dialog for loading from Nexus"""
        try:
            vicon = nexus.viconnexus()
        except GaitDataError as e:
            qt_message_dialog('Error: %s' % str(e))
            return
        self._load_dialog(vicon)

    def _load_dialog_c3d(self):
        """Dialog for loading from c3d"""
        fout = QtWidgets.QFileDialog.getOpenFileName(self, 'Open C3D file',
                                                     '',
                                                     u'C3D files (*.c3d)')[0]
        if fout:
            self._load_dialog(fout)

    def _emg_filter_dialog(self):
        dlg = EMGFilterDialog(self.emg_passband)
        if not dlg.exec_():
            return
        else:
            self._reset_emg_filter(dlg.spEMGLow.value(), dlg.spEMGHigh.value())

    def _load_dialog(self, source):
        """Dialog for loading data """
        dlg = LoadDialog()
        if not dlg.exec_():
            return
        side = 'R' if dlg.rbRight.isChecked() else 'L'
        # prepend side to configured EMG channel names
        emg_chs = [side+ch for ch in cfg.tardieu.emg_chs]
        ang0_nexus = dlg.spNormAngle.value()

        self._nonresp()
        try:
            self._tardieu_plot.load_data(source, emg_chs, self.emg_passband,
                                         ang0_nexus)
        except GaitDataError as e:
            qt_message_dialog('Error: %s' % str(e))
            self._resp()
            return

        self._tardieu_plot.plot_data()
        self.canvas.draw()
        self.canvas.setFocus()
        self._update_status()
        self._update_marker_status()
        # set data controls to match what was loaded
        self.spEMGYScale.setValue(cfg.plot.emg_yscale*1e3)  # mV
        # enable all controls
        self._resp()
        self._set_data_controls(True)

    def _save_plot(self):
        """Save a pdf report"""
        fn_pdf = self._tardieu_plot.trial.trialname + '.pdf'
        fname = QtWidgets.QFileDialog.getSaveFileName(self, 'Save plot',
                                                      fn_pdf,
                                                      u'PDF files (*pdf)')[0]
        if not fname:
            return
        try:
            with PdfPages(fname) as pdf:
                page_size = (11.69, 8.27)
                # create header page
                #timestr = time.strftime("%d.%m.%Y")
                fig_hdr = Figure()
                FigureCanvas(fig_hdr)
                ax = fig_hdr.add_subplot(111)
                ax.set_axis_off()
                txt = self._tardieu_plot.status_text
                ax.text(.5, .8, txt, ha='center', va='center', weight='bold',
                        fontsize=12)
                fig_hdr.set_size_inches(page_size[0], page_size[1])
                pdf.savefig(fig_hdr)
                figx, data_axes, legend_ax = self._tardieu_plot.plot_data(interactive=False,
                                                                          emg_yscale=self.spEMGYScale.value())
                FigureCanvas(figx)
                figx.set_size_inches(page_size[0], page_size[1])
                # full view
                self._tardieu_plot.markers.plot_on_axes(data_axes)
                self._tardieu_plot.markers.legend(legend_ax)
                pdf.savefig(figx)
                # create zoomed version
                fast_rng = self._tardieu_plot._get_fast_movement()
                for ax in data_axes:
                    ax.set_xlim(fast_rng[0], fast_rng[1])
                pdf.savefig(figx)
        except IOError:
            qt_message_dialog('Error writing %s, '
                              'check that file is not already open.' % fname)
            return

    def _update_status(self):
        """Update the status text"""
        status = self._tardieu_plot.status_text
        self.lblStatus.setText(status)

    def _update_marker_status(self):
        """Update the marker status text"""
        status = self._tardieu_plot.marker_status_text
        self.lblMarkerStatus.setText(status)

    def _clear_markers(self):
        """Clear all markers"""
        if self._tardieu_plot.markers is not None:
            self._tardieu_plot.markers.clear()
            self.canvas.draw()
            self._update_marker_status()

    def closeEvent(self, event):
        """Confirm and close application"""
        reply = qt_yesno_dialog('Are you sure?')
        if reply == QtWidgets.QMessageBox.YesRole:
            event.accept()
        else:
            event.ignore()


class Markers(object):
    """ Manage vertical marker lines at multiple axes.
    The markers are created as matplotlib axvline()s. """

    def __init__(self, marker_colors, marker_width, axes):
        """ Initialize.
        marker_colors: the colors (and max. number) of markers
        marker_width: the line width for the markers
        axes: all axes to put the markers in """
        self._markers = OrderedDict()  # markers are keyed by x coordinate
        self._axes = axes
        self.marker_colors = marker_colors
        self.marker_width = marker_width
        self.max_markers = len(self.marker_colors)

    def add_on_click(self, x):
        """Add a marker on mouse click"""
        if x not in self._markers.keys():
            if len(self._markers) == self.max_markers:
                qt_message_dialog('You can place a maximum of %d markers' %
                                  self.max_markers)
            else:
                self.add(x)

    def add(self, x, annotation=''):
        """Add a marker at point x with optional annotation"""
        if x in self._markers.keys():  # marker already exists here
            return
        else:
            cols_in_use = [m['color'] for m in self._markers.values()]
            col = list((set(self.marker_colors) - set(cols_in_use)))[0]
            self._markers[x] = dict()
            self._markers[x]['annotation'] = annotation
            self._markers[x]['color'] = col
            # each axis gets its own line artist
            for ax in self._axes:
                self._markers[x][ax] = ax.axvline(x=x, color=col,
                                                  linewidth=self.marker_width)
                # generate picker events at given tolerance
                self._markers[x][ax].set_picker(3)

    def delete(self, x):
        """Delete marker by location"""
        for ax in self._axes:
            self._markers[x][ax].remove()
        self._markers.pop(x)

    def delete_artist(self, artist, ax):
        """Delete marker by artist at axis ax"""
        # need to find the marker that has the corresponding artist
        for x, m in self._markers.items():
            if m[ax] == artist:
                self.delete(x)
                return

    def clear(self):
        """Remove all markers"""
        for marker in self._markers:
            self.delete(marker)

    def plot_on_axes(self, axes):
        """Plot all markers on given axes (iterable)"""
        for ax in axes:
            for x, m in self._markers.items():
                ax.axvline(x=x, color=m['color'], linewidth=self.marker_width)

    def legend(self, ax, ncol=2):
        """Create legend"""
        artists = list()
        legtxts = list()
        for mkr, anno, col in self.marker_info:
            artists.append(matplotlib.lines.Line2D((0, 1), (0, 0),
                                                   linewidth=self.marker_width,
                                                   color=col))
            legtxts.append(u'%.3f s: %s' % (mkr, anno))
        ax.legend(artists, legtxts, loc='upper left', ncol=ncol,
                  prop={'size': cfg.plot.label_fontsize})

    @property
    def marker_info(self):
        """Return tuple of marker, annotation, and color"""
        annotations = [m['annotation'] for m in self._markers.values()]
        cols_in_use = [m['color'] for m in self._markers.values()]
        return zip(self._markers.keys(), annotations, cols_in_use)


class TardieuPlot(object):
    """ Create matplotlib graphs for Tardieu analysis """

    def __init__(self):
        """Initialize but do not plot anything yet"""
        # adjustable params
        # TODO: some could go into config
        self.marker_button = 1  # mouse button for placing markers
        self.marker_del_button = 3  # remove marker
        self.marker_key = 'shift'  # modifier key for markers
        self.markers = None
        # FIXME: check colors
        self.marker_colors = ['orange', 'green', 'red', 'brown',
                              'gray', 'purple']
        self.marker_width = 1.5
        self.width_ratio = [1, 5]
        self.text_fontsize = 9
        self.margin = .025  # margin at edge of plots
        self.margin = 0
        self.narrow = False
        self.hspace = .4
        self.wspace = .5
        self.emg_automark_chs = ['Gas', 'Sol']   # FIXME: into config?
        self.data_axes = list()  # axes that actually contain data
        self.emg_axes = list()
        # these are callbacks that should be registered by the creator
        self._update_marker_status = None
        self._update_status = None
        self.fig = Figure()

    def load_data(self, source, emg_chs, emg_passband, ang0_nexus):
        """Load the Tardieu data.
        emg_chs: list of EMG channel names to use

        Returns True on successful data load, False otherwise
        """

        self.trial = Trial(source)
        self.trial.emg.passband = emg_passband

        # the 'true' physiological starting angle (given as a param)
        self.ang0_nexus = ang0_nexus
        self.emg_chs = emg_chs
        self.time = self.trial.t / self.trial.framerate  # time axis in sec
        self.tmax = self.time[-1]
        self.nframes = len(self.time)

        # read EMG data
        self.emgdata = dict()
        self.emg_rms = dict()
        for ch in self.emg_chs:
            t_, self.emgdata[ch] = self.trial.get_emg_data(ch)
            self.emg_rms[ch] = rms(self.emgdata[ch], cfg.emg.rms_win)

        if cfg.tardieu.acc_chs:
            accdata_ = self.trial.accelerometer_data['data']
            # compute vector sum over given accelerometer chs
            try:
                accsigs = [accdata_[ch] for ch in cfg.tardieu.acc_chs]
            except KeyError:
                raise GaitDataError('No such accelerometer channel %s' % ch)
            acctot = np.stack(accsigs)
            self.acctot = np.sqrt(np.sum(acctot**2, 0))
        else:
            self.acctot = None

        # FIXME: self.time?
        self.time_analog = t_ / self.trial.analograte

        # read marker data and compute segment angle
        mnames = cfg.tardieu.marker_names
        data = read_data.get_marker_data(source, mnames)

        P0 = data[mnames[0]]
        P1 = data[mnames[1]]
        P2 = data[mnames[2]]
        # stack so that marker changes along 2nd dim for segment_angles
        Pall = np.stack([P0, P1, P2], axis=1)
        # compute segment angles (deg)
        self.angd = segment_angles(Pall) / np.pi * 180
        # this is our calculated starting angle
        ang0_our = np.median(self.angd[~np.isnan(self.angd)][:10])
        # normalize: plantarflexion negative, our starting angle equals
        # the starting angle given in Nexus (i.e. if ang0_nexus is 95,
        # we offset the data to start at -5 deg)
        # if starting angle is not specified in Nexus, it defaults to 90 deg
        self.angd = 90 - self.ang0_nexus - self.angd + ang0_our
        return True

    def plot_data(self, interactive=True, emg_yscale=None):
        """ Plot the data. Can plot either on the main (interactive) display
        or a new mpl Figure() (which will be returned)"""

        fig = self.fig if interactive else Figure()
        fig.clear()
        data_axes = list()
        if interactive:  # save trace objects for later modification by GUI
            self.emg_traces, self.rms_traces = dict(), dict()

        nrows = len(self.emg_chs) + 3   # emgs + acc + marker data
        if self.acctot is not None:
            nrows += 1

        # add one row for legend if not in interactive mode
        if not interactive:
            hr = [1] * nrows + [.5]
            nrows += 1
            gs = gridspec.GridSpec(nrows, 1, height_ratios=hr)
            legend_ax = fig.add_subplot(gs[-1, 0])
            legend_ax.set_axis_off()
        else:
            gs = gridspec.GridSpec(nrows, 1)
            legend_ax = None

        # EMG plots
        ind = 0
        for ch in self.emg_chs:
            sharex = None if ind == 0 or not interactive else data_axes[0]
            ax = fig.add_subplot(gs[ind, 0], sharex=sharex)
            emgtr_, = ax.plot(self.time_analog, self.emgdata[ch]*1e3,
                              linewidth=cfg.plot.emg_linewidth)
            rmstr_, = ax.plot(self.time_analog, self.emg_rms[ch]*1e3,
                              linewidth=cfg.plot.emg_rms_linewidth,
                              color='black')
            data_axes.append(ax)
            if interactive:
                self.emg_traces[ind] = emgtr_
                self.rms_traces[ind] = rmstr_
                self.emg_axes.append(ax)

            ysc = (cfg.plot.emg_yscale if emg_yscale is None
                   else emg_yscale / 1.e3)
            ax.set_ylim([-ysc*1e3, ysc*1e3])
            ax.set(ylabel='mV')
            ax.set_title(ch)
            ind += 1

        # total acceleration
        if self.acctot is not None:
            sharex = None if ind == 0 or not interactive else data_axes[0]
            ax = fig.add_subplot(gs[ind, 0], sharex=sharex)
            ax.plot(self.time_analog, self.acctot,
                    linewidth=cfg.plot.emg_linewidth)
            # FIXME: no calibration yet so data is assumed to be in mV
            # ax.set(ylabel=u'm/s²')
            ax.set(ylabel=u'mV')
            ax.set_title('Accelerometer vector sum')
            data_axes.append(ax)
            ind += 1

        # angle plot
        sharex = None if ind == 0 or not interactive else data_axes[0]
        ax = fig.add_subplot(gs[ind, 0], sharex=sharex)
        ax.plot(self.time, self.angd, linewidth=cfg.plot.model_linewidth)
        ax.set(ylabel='deg')
        ax.set_title('Angle')
        data_axes.append(ax)
        ind += 1

        # angular velocity plot
        sharex = None if ind == 0 or not interactive else data_axes[0]
        ax = fig.add_subplot(gs[ind, 0], sharex=sharex)
        self.angveld = self.trial.framerate * np.diff(self.angd, axis=0)
        ax.plot(self.time[:-1], self.angveld,
                linewidth=cfg.plot.model_linewidth)
        ax.set(ylabel='deg/s')
        ax.set_title('Angular velocity')
        data_axes.append(ax)
        ind += 1

        # angular acceleration plot
        sharex = None if ind == 0 or not interactive else data_axes[0]
        ax = fig.add_subplot(gs[ind, 0], sharex=sharex)
        self.angaccd = np.diff(self.angveld, axis=0)
        ax.plot(self.time[:-2], self.angaccd,
                linewidth=cfg.plot.model_linewidth)
        ax.set(xlabel='Time (s)', ylabel=u'deg/s²')
        ax.set_title('Angular acceleration')
        data_axes.append(ax)

        for ax in data_axes:
            self._adj_fonts(ax)

        if interactive:
            self.data_axes = data_axes
            self.tmin, self.tmax = self.data_axes[0].get_xlim()
            # create markers
            markers = Markers(self.marker_colors, self.marker_width,
                              self.data_axes)

            # place the auto markers
            tmin_ = max(self.time[0], self.tmin)
            tmax_ = min(self.time[-1], self.tmax)
            fmin, fmax = self._time_to_frame([tmin_, tmax_],
                                             self.trial.framerate)
            smin, smax = self._time_to_frame([tmin_, tmax_],
                                             self.trial.analograte)
            # max. velocity
            velr = self.angveld[fmin:fmax]
            velmaxind = np.nanargmax(velr) / self.trial.framerate + tmin_
            markers.add(velmaxind, annotation='Max. velocity')
            # min. acceleration
            accr = self.angaccd[fmin:fmax]
            accmaxind = np.nanargmin(accr) / self.trial.framerate + tmin_
            markers.add(accmaxind, annotation='Min. acceleration')

            for ch in self.emg_chs:
                # check if channel is tagged for automark
                if any([s in ch for s in self.emg_automark_chs]):
                    rmsdata = self.emg_rms[ch][smin:smax]
                    rmsmaxind = np.argmax(rmsdata) / self.trial.analograte + tmin_
                    markers.add(rmsmaxind, annotation='%s max. RMS' % ch)

            # connect callbacks
            for ax in self.data_axes:
                ax.callbacks.connect('xlim_changed', self._xlim_changed)
            fig.canvas.mpl_connect('button_press_event', self._onclick)
            # catch key press
            # fig.canvas.mpl_connect('key_press_event', self._onpress)
            # pick handler
            fig.canvas.mpl_connect('pick_event', self._onpick)
            #
            self._last_click_event = None
            self.markers = markers

        fig.set_tight_layout(True)
        return fig, data_axes, legend_ax

    def _rescale_emg(self, yscale):
        """Takes new EMG yscale in mV"""
        for ax in self.emg_axes:
            ax.set_ylim(-yscale, yscale)

    def _get_fast_movement(self):
        """Get x range around fast movement"""
        velmaxt = self.time[np.nanargmax(self.angveld)]
        return velmaxt-.5, velmaxt+1.5

    def _xzoom(self, x1, x2):
        for ax in self.data_axes:
            ax.set_xlim(x1, x2)

    def _xzoom_to_fast(self):
        """Zoom x to fast movement"""
        rng = self._get_fast_movement()
        self._xzoom(rng[0], rng[1])

    def _xzoom_reset(self):
        self._xzoom(self.time[0], self.time[-1])

    def _reset_emg_filter(self, f1, f2):
        """Takes new EMG lowpass and highpass values"""
        logger.debug('reset EMG filter to %.2f-%.2f' % (f1, f2))
        self.trial.emg.passband = [f1, f2]
        for ind, ch in enumerate(self.emg_chs):
            t_, self.emgdata[ch] = self.trial.get_emg_data(ch)
            self.emg_rms[ch] = rms(self.emgdata[ch], cfg.emg.rms_win)
            self.emg_traces[ind].set_ydata(self.emgdata[ch]*1e3)
            self.rms_traces[ind].set_ydata(self.emg_rms[ch]*1e3)

    @staticmethod
    def _adj_fonts(ax):
        """Adjust font sizes on an axis"""
        ax.xaxis.label.set_fontsize(cfg.plot.label_fontsize)
        ax.yaxis.label.set_fontsize(cfg.plot.label_fontsize)
        ax.title.set_fontsize(cfg.plot.title_fontsize)
        ax.tick_params(axis='both', which='major',
                       labelsize=cfg.plot.ticks_fontsize)

    @staticmethod
    def _time_to_frame(times, rate):
        """Convert time to samples (according to rate)"""
        return np.round(rate * np.array(times)).astype(int)

    def tight_layout(self):
        """ Auto set spacing between/around axes """
        self.fig.set_tight_layout(True)
        # not sure if works/needed
        # self.gs.update(hspace=self.hspace, wspace=self.wspace,
        #               left=self.margin, right=1-self.margin)
        # probably needed
        # self.fig.canvas.draw()

    def _xlim_changed(self, ax):
        """Callback for x limits change, e.g. on zoom"""
        # we need to get the limits from the axis that was zoomed
        # (the limits are not instantly propagated by sharex)
        self.tmin, self.tmax = ax.get_xlim()
        self.fig.canvas.draw()
        self._update_status()

    def _toggle_narrow_callback(self, event):
        """Toggle narrow/wide display"""
        self.narrow = not self.narrow
        wratios = [1, 1] if self.narrow else self.width_ratio
        btext = 'Wide view' if self.narrow else 'Narrow view'
        self.gs.set_width_ratios(wratios)
        self._narrowbutton.label.set_text(btext)
        self.gs.update()
        # FIXME: canvas ref
        self.fig.canvas.draw()

    def _onpick(self, event):
        """Gets triggered on pick event, i.e. selection of existing marker"""
        if self._toolbar.mode:  # do not respond if toolbar buttons enabled
            return
        mevent = event.mouseevent
        # prevent handling the same onpick event multiple times (e.g.
        # if multiple markers get picked on a single click)
        if self._last_click_event == mevent:
            return
        if (mevent.button != self.marker_del_button or
           mevent.key != self.marker_key):
            return
        self.markers.delete_artist(event.artist, mevent.inaxes)
        self._last_click_event = mevent
        self.fig.canvas.draw()
        self._update_marker_status()

    def _onpress(self, event):
        """Keyboard event handler"""
        if event.key == 'tab':
            self._toggle_narrow_callback(event)

    def _onclick(self, event):
        """Gets triggered by a mouse click on canvas"""
        if self._toolbar.mode:  # do not respond if toolbar buttons are enabled
            return
        if event.inaxes not in self.data_axes:
            return
        # prevent handling the same click event multiple times
        # onpick and onclick may get triggered simultaneously
        if event == self._last_click_event:
            return
        if event.button != self.marker_button or event.key != self.marker_key:
            return
        x = event.xdata
        self.markers.add_on_click(x)
        self._last_click_event = event
        self.fig.canvas.draw()
        self._update_marker_status()

    @property
    def marker_status_text(self):
        """Return marker status text in HTML"""
        s = u''
        # each marker gets text of its own color
        for marker, anno, col in self.markers.marker_info:
            frame = self._time_to_frame(marker, self.trial.framerate)
            s += u"<font color='%s'>" % col
            # len of acc signal is nframes - 2
            if frame < 0 or frame >= self.nframes - 2:
                s += u'Marker outside data range'
            else:
                s += u'Marker @%.3f s' % marker
                s += (' (%s):<br>') % anno if anno else ':<br>'
                s += u'dflex: %.2f° vel: %.2f°/s' % (self.angd[frame],
                                                       self.angveld[frame])
                s += u' acc: %.2f°/s²<br><br>' % self.angaccd[frame]
            s += u'</font>'
        return s

    @property
    def status_text(self):
        """Create the status text"""

        # find the limits of the data that is shown
        tmin_ = max(self.time[0], self.tmin)
        tmax_ = min(self.time[-1], self.tmax)
        s = u'Trial name: %s\n' % self.trial.trialname
        s += u'Description: %s\n' % (self.trial.eclipse_data['DESCRIPTION'])
        s += u'Notes: %s\n' % (self.trial.eclipse_data['NOTES'])
        s += u'Angle offset: '
        s += (u' %.1f°\n' % self.ang0_nexus) if self.ang0_nexus else u'none\n'
        s += u'Data range shown: %.2f - %.2f s\n' % (tmin_, tmax_)
        # frame indices corresponding to time limits
        fmin, fmax = self._time_to_frame([tmin_, tmax_], self.trial.framerate)
        if fmin == fmax:
            s += 'Zoomed in to a single frame\nPlease zoom out for info'
            return s
        else:
            smin, smax = self._time_to_frame([tmin_, tmax_],
                                             self.trial.analograte)
            s += u'In frames: %d - %d\n\n' % (fmin, fmax)
            # foot angle in chosen range and the maximum
            angr = self.angd[fmin:fmax]
            # check if we zoomed to all-nan region of angle data
            if np.all(np.isnan(angr)):
                s += 'No valid data in region'
                return s
            angmax = np.nanmax(angr)
            angmaxind = np.nanargmax(angr) / self.trial.framerate + tmin_
            angmin = np.nanmin(angr)
            angminind = np.nanargmin(angr) / self.trial.framerate + tmin_
            # same for velocity
            velr = self.angveld[fmin:fmax]
            velmax = np.nanmax(velr)
            velmaxind = np.nanargmax(velr) / self.trial.framerate + tmin_
            s += u'Values for range shown:\n'
            s += u'Max. dorsiflexion: %.2f° @ %.2f s\n' % (angmax, angmaxind)
            s += u'Max. plantarflexion: %.2f° @ %.2f s\n' % (angmin, angminind)
            s += u'Max velocity: %.2f°/s @ %.2f s\n' % (velmax, velmaxind)
            for ch in self.emg_chs:
                rmsdata = self.emg_rms[ch][smin:smax]
                rmsmax = rmsdata.max()
                rmsmaxind = np.argmax(rmsdata) / self.trial.analograte + tmin_
                s += u'%s max RMS: %.2f mV @ %.2f s\n' % (ch, rmsmax*1e3,
                                                          rmsmaxind)
            return s


def main():

    logging.basicConfig(level=logging.DEBUG)

    # uiparser logger makes too much noise
    logging.getLogger('PyQt5.uic').setLevel(logging.WARNING)

    app = QtWidgets.QApplication(sys.argv)
    win = TardieuWindow()
    win.show()
    sys.exit(app.exec_())



# -*- coding: utf-8 -*-
"""
Session related functions


@author: Jussi (jnu@iki.fi)
"""

from builtins import str
import io
import os.path as op
import json
import datetime
import glob
import re
import logging

from .eclipse import get_eclipse_keys
from .videos import get_trial_videos
from .envutils import GaitDataError
from .config import cfg


logger = logging.getLogger(__name__)

json_keys = ['fullname', 'hetu', 'session_description', 'report_notes']


def default_info():
    """Return info dict with placeholder values"""
    return {key: None for key in json_keys}


def load_info(session):
    """Return the patient info dict from the given session"""
    fname = op.join(session, 'patient_info.json')
    if op.isfile(fname):
        with io.open(fname, 'r', encoding='utf-8') as f:
            try:
                info = json.load(f)
                extra_keys = set(info.keys()) - set(json_keys)
                if extra_keys:
                    logger.warning('Extra keys %s in patient info file %s'
                                   % (extra_keys, fname))
                    for key in extra_keys:
                        info.pop(key)
                missing_keys = set(json_keys) - set(info.keys())
                if missing_keys:
                    logger.warning('Missing keys %s in patient info file %s'
                                   % (missing_keys, fname))
                    # supply default values for missing keys
                    for key in missing_keys:
                        info[key] = default_info()[key]
            except (UnicodeDecodeError, EOFError, IOError, TypeError,
                    ValueError):
                raise GaitDataError('Error loading patient info file %s'
                                    % fname)
    else:
        info = None

    return info


def save_info(session, patient_info):
    """Save patient info."""
    fname = op.join(session, 'patient_info.json')
    try:
        with io.open(fname, 'w', encoding='utf-8') as f:
            f.write(str(json.dumps(patient_info, ensure_ascii=False)))
    except (UnicodeDecodeError, EOFError, IOError, TypeError):
        raise GaitDataError('Error saving patient info file %s ' % fname)


def _merge_session_info(sessions):
    """merge patient info files across sessions. fullname and hetu must match.
    Returns dict of individual session infos and the mergeD info"""
    session_infos = {session: (load_info(session) or default_info())
                     for session in sessions}
    info = default_info()
    # ignore the session description
    for key in ['fullname', 'hetu', 'report_notes']:
        allvals = set([session_infos[session][key] for session in sessions])
        if None in allvals:
            allvals.remove(None)
        if key == 'fullname' or key == 'hetu':
            if len(allvals) > 1:
                logger.warning('name / hetu do not match across sessions')
                return session_infos, None
        if key == 'report_notes':
            if len(allvals) > 1:
                logger.warning('report notes do not match across sessions')
        # in case of conflicts, this just picks from the last session
        info[key] = allvals.pop() if allvals else None
    return session_infos, info


def _enf2other(fname, ext):
    """Converts name of trial .enf file to corresponding .c3d or other
    file type"""
    enfre = '\.*.Trial\d*.enf'  # .Trial followed by zero or more digits
    res = re.search(enfre, fname)
    if res is None:
        raise GaitDataError('Filename %s is not a trial .enf' % fname)
    return fname.replace(res.group(), '.%s' % ext)


def get_session_date(sessionpath):
    """Return date when session was recorded (datetime.datetime object)"""
    enfs = get_session_enfs(sessionpath)
    x1ds = [_enf2other(fn, 'x1d') for fn in enfs]
    if not x1ds:
        raise GaitDataError('Invalid session %s' % sessionpath)
    else:
        x1d = x1ds[0]
        if not op.isfile(x1d):
            raise GaitDataError('x1d trial file %s does not exist' % x1d)
        return datetime.datetime.fromtimestamp(op.getmtime(x1d))


def get_session_enfs(sessionpath):
    """Return list of .enf files for the session """
    enfglob = op.join(sessionpath, '*Trial*.enf')
    for enf in glob.iglob(enfglob):
        yield enf


def _filter_by_eclipse_keys(enfs, patterns, eclipse_keys):
    """Filter for enfs whose Eclipse key values match given patterns."""
    if not isinstance(patterns, list):
        patterns = [patterns]
    if not isinstance(eclipse_keys, list):
        eclipse_keys = [eclipse_keys]
    for enf in enfs:
        ecldi = {key.upper(): val.upper() for key, val in
                 get_eclipse_keys(enf).items()}
        eclvals = [val for key, val in ecldi.items() if key in eclipse_keys]
        for pattern in patterns:
            if any(pattern.upper() in eclval for eclval in eclvals):
                yield enf


def _filter_by_tags(enfs, tags):
    """Filter by given tags"""
    return _filter_by_eclipse_keys(enfs, tags, cfg.eclipse.tag_keys)


def _filter_by_type(enfs, trial_type):
    """Filter by trial type"""
    return _filter_by_eclipse_keys(enfs, trial_type, 'TYPE')


def _filter_to_c3ds(enfs):
    """Convert enf filenames to c3d"""
    for enf in enfs:
        yield _enf2other(enf, 'c3d')


def _filter_exists(files):
    for f in files:
        if op.isfile(f):
            yield f


def get_c3ds(sessionpath, tags=None, trial_type=None, check_if_exists=False):
    """Get specified c3d files for session."""
    enfs = get_session_enfs(sessionpath)
    if trial_type is not None:
        enfs = _filter_by_type(enfs, trial_type)
    if tags is not None:
        enfs = _filter_by_tags(enfs, tags)
    c3ds = _filter_to_c3ds(enfs)
    if check_if_exists:
        c3ds = _filter_exists(c3ds)
    return list(c3ds)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Created on Thu Sep 03 14:54:34 2015

EMG consistency plot from Nexus. Automatically picks trials based on Eclipse
description and defined search strings.

@author: Jussi (jnu@iki.fi)
"""

import logging
import argparse
import os.path as op
from itertools import cycle

from gaitutils import (Plotter, cfg, register_gui_exception_handler, EMG,
                       GaitDataError, sessionutils, nexus, Trial, plot_plotly)
from gaitutils.layouts import rm_dead_channels_multitrial

logger = logging.getLogger(__name__)


def do_plot(sessionpath=None, tags=None, show=True, make_pdf=True,
            backend=None):

    if sessionpath is None:
        sessionpath = nexus.get_sessionpath()

    if backend is None:
        backend = cfg.plot.backend

    c3dfiles = sessionutils.get_c3ds(sessionpath, tags=cfg.eclipse.tags,
                                     trial_type='dynamic')

    if not c3dfiles:
        raise GaitDataError('No marked trials found for current session')

    linecolors = cfg.plot.overlay_colors
    ccolors = cycle(linecolors)

    layout = cfg.layouts.overlay_std_emg
    emgs = [EMG(tr) for tr in c3dfiles]
    layout = rm_dead_channels_multitrial(emgs, layout)

    if backend == 'matplotlib':
        pl = Plotter()
        pl.layout = layout

        for i, trialpath in enumerate(c3dfiles):
            if i > len(linecolors):
                logger.warning('not enough colors for plot!')
            pl.open_trial(c3dfiles[i])

            emg_active = any([pl.trial.emg.status_ok(ch) for ch in
                              cfg.emg.channel_labels])
            if not emg_active:
                continue

            plot_emg_normaldata = (trialpath == c3dfiles[-1])

            pl.plot_trial(emg_tracecolor=next(ccolors),
                          maintitle='', annotate_emg=False,
                          superpose=True, show=False,
                          plot_emg_normaldata=plot_emg_normaldata)

        if not pl.fig:
            raise GaitDataError('None of the trials have valid EMG data')

        maintitle = ('EMG consistency plot, '
                     'session %s' % pl.trial.sessiondir)
        pl.set_title(maintitle)

        if show:
            pl.show()

        if make_pdf:
            pl.create_pdf('emg_consistency.pdf')

        return pl.fig

    elif backend == 'plotly':

        trials = [Trial(c3d) for c3d in c3dfiles]
        maintitle = ('EMG consistency plot, session %s' %
                     op.split(sessionpath)[-1])
        plot_plotly.plot_trials_browser(trials, layout,
                                        legend_type='short_name_with_tag',
                                        maintitle=None)


if __name__ == '__main__':

    parser = argparse.ArgumentParser()
    parser.add_argument('--tags', metavar='p', type=str, nargs='+',
                        help='strings that must appear in trial '
                        'description or notes')
    args = parser.parse_args()
    logging.basicConfig(level=logging.DEBUG)
    register_gui_exception_handler()
    do_plot(tags=args.tags)



# -*- coding: utf-8 -*-
"""
Widget plugin for Qt Designer

@author: Jussi (jnu@iki.fi)
"""
from __future__ import absolute_import

from qt_widgets import NiceListWidget
from PyQt5 import QtDesigner


class NiceListWidgetPlugin(QtDesigner.QPyDesignerCustomWidgetPlugin):

    def __init__(self, parent=None):

        QtDesigner.QPyDesignerCustomWidgetPlugin.__init__(self)
        self.initialized = False

    def initialize(self, core):

        if self.initialized:
            return

        self.initialized = True

    def isInitialized(self):

        return self.initialized

    def createWidget(self, parent):
        return NiceListWidget(parent)

    def name(self):
        return "NiceListWidget"

    def group(self):
        return "Custom Widgets"

    def isContainer(self):
        return False

    def includeFile(self):
        return "gaitutils.gui.qt_widgets"



# -*- coding: utf-8 -*-
"""
Handles gaitutils config files.

@author: Jussi (jnu@iki.fi)
"""
from __future__ import print_function

from builtins import str
from builtins import object
import configparser
import ast
import os.path as op
import os
import copy
import sys
import subprocess
from pkg_resources import resource_filename

from . import envutils


""" Work around stdout and stderr not being available, if we are run
using pythonw.exe on Windows. Without this, exception will be raised
e.g. on any print statement. """
if (sys.platform.find('win') != -1 and sys.executable.find('pythonw') != -1 and
   not envutils.run_from_ipython()):
    blackhole = open(os.devnull, 'w')
    sys.stdout = sys.stderr = blackhole

# default config
cfg_template = resource_filename(__name__, 'data/default.cfg')
# user specific config
# On Windows, this typically puts the config at C:\Users\Username, since the
# USERPROFILE environment variable points there. Putting the config in a
# networked home dir requires some tinkering with environment variables
# (e.g. setting HOME)
homedir = op.expanduser('~')
cfg_user = op.join(homedir, '.gaitutils.cfg')


class Section(object):
    """ A config section """

    def __init__(self, di):
        """ The config items for the section are provided as key/value dict,
        where values are strings. """
        self._dict = di

    def __getattr__(self, item):
        """ Implements attribute access, i.e. section.item. The items don't
        exist as instance variables, so referencing them will cause
        __getattr__ to be called. The items are returned from the section dict
        and automatically converted from strings to Python types. """
        if item not in self._dict:
            raise AttributeError('Config key does not exist: %s' % item)
        try:
            return ast.literal_eval(self._dict[item])
        except ValueError:
            raise ValueError('Could not convert value %s to Python type' %
                             self._dict[item])


class EpicParser(configparser.SafeConfigParser):
    """ Extends SafeConfigParser by providing convenient attribute access
    (by parser.section.item) and autoconversion from strings to Python
    types.
    For modifying config items, it's still necessary to use the
    syntax parser[section][item] = data, where data must always be a string
    type. """

    def __getitem__(self, section):
        """ Returns the parser section dictionary. """
        return self._sections[section]

    def __repr__(self):
        return '<EpicParser>'

    def __getattr__(self, section):
        """ Implements attribute access, i.e. parser.section or more commonly
        parser.section.item. A new Section instance is created from the parser
        data. """
        return Section(self._sections[section])

    def write_file(self, filename):
        """ Save config into file """
        # using the with statement would force-close the file even in case
        # of errors, which could lead to broken (e.g. empty) files
        fh = open(filename, 'wt')
        cfg.write(fh)
        fh.close()

    def load_default(self):
        """ Load default config """
        self.read(cfg_template)


# provide the global cfg instance
# read template config
cfg = EpicParser()
cfg.load_default()
cfg_tpl_di = copy.deepcopy(cfg._sections)  # save the template config

# read user config
if not op.isfile(cfg_user):
    print('no config file, trying to create %s' % cfg_user)
    cfg.write_file(cfg_user)
else:
    print('reading user config from %s' % cfg_user)
    cfg.read(cfg_user)

# check for extra entries in user config
no_check = ['layouts']
cfg_user_di = cfg._sections
for sname, section_ in cfg_user_di.items():
    if sname not in cfg_tpl_di:
        print('WARNING: unused (deprecated?) section %s in user config'
              % sname)
    else:
        if sname not in no_check:
            for key in section_:
                if key not in cfg_tpl_di[sname]:
                    print('WARNING: unused (deprecated?) key '
                          '%s.%s in user config' % (sname, key))

# handle some deprecated/changed types for user convenience
if not isinstance(cfg.plot.emg_yscale, float):
    ysc = cfg.plot.emg_yscale[1]
    print('WARNING: emg_yscale was changed to float, using %g' % ysc)
    cfg['plot']['emg_yscale'] = str(cfg.plot.emg_yscale[1])

if cfg.general.normaldata_files == 'default':
    fn = resource_filename('gaitutils', 'data/normal.gcd')
    cfg['general']['normaldata_files'] = repr([fn])

if cfg.general.videoconv_path == 'default':
    fn = resource_filename('gaitutils', 'thirdparty/ffmpeg2theora.exe')
    cfg['general']['videoconv_path'] = repr(fn)


sys.stdout.flush()  # make sure that warnings are printed out



# -*- coding: utf-8 -*-
"""
Compute statistics across/within trials


@author: Jussi (jnu@iki.fi)
"""

import logging
import numpy as np

from .trial import Trial, Gaitcycle
from .envutils import GaitDataError
from . import models

logger = logging.getLogger(__name__)


class AvgTrial(Trial):
    """ Trial containing cycle-averaged data, for use with plotter
    TODO: does not support legends yet """

    def __init__(self, c3dfiles, fp_cycles_only=False, max_dist=None):
        avgdata, stddata, n_ok, _ = average_trials(c3dfiles, max_dist=max_dist,
                                                   fp_cycles_only=fp_cycles_only)
        # nfiles may be misleading since not all trials may contain valid data
        self.nfiles = len(c3dfiles)
        self.trialname = 'Averages from %d trials' % self.nfiles
        self.source = 'averaged data'
        self.name = 'Unknown'
        self._model_data = avgdata
        self.stddev_data = stddata
        self.n_ok = n_ok
        self.t = np.arange(101)  # 0..100%
        # fake 2 gait cycles, L/R
        self.cycles = list()
        self.cycles.append(Gaitcycle(0, 101, 60, 'R', True, 1, 1000,
                                     name='Right average', trial=self))
        self.cycles.append(Gaitcycle(0, 101, 60, 'L', True, 1, 1000,
                                     name='Left average', trial=self))
        self.ncycles = 2
        self.sessionpath = None

    def get_model_data(self, var):
        return self.t, self._model_data[var]

    def set_norm_cycle(self, cycle=None):
        if cycle is None:
            raise ValueError('AvgTrial does not support unnormalized data')
        else:
            logger.debug('setting norm. cycle for AvgTrial (no effect)')


def average_trials(trials, max_dist=None, fp_cycles_only=False,
                   reject_zeros=True):
    """ Average model data from several trials.

    trials: list
        filename, or list of filenames (c3d) to read trials from, or list
        of Trial instances
    max_dist: maximum curve distance from median, for outlier rejection
    fp_cycles_only: bool
        If True, only collect data from forceplate cycles. Kinetics will always
        be collected from forceplate cycles only.

    Returns:

    avgdata: dict
        The data
    stddata: dict
        Standard dev for each var
    N_ok: dict
        N of accepted cycles for each var
    Ncyc: dict
        WIP
    """
    data, Ncyc = _collect_model_data(trials, fp_cycles_only=fp_cycles_only)

    stddata = dict()
    avgdata = dict()
    N_ok = dict()

    for var, vardata in data.items():
        if vardata is None:
            stddata[var] = None
            avgdata[var] = None
            N_ok[var] = 0
            continue
        else:
            Ntot = vardata.shape[0]

            # drop outliers
            if max_dist is not None:
                outliers = _outlier_rows(vardata, max_dist)
                N_out = np.count_nonzero(outliers)
                if N_out > 0:
                    logger.debug('%s: dropping %d outlier curves' %
                                 (var, N_out))
                    vardata = vardata[~outliers, :] if N_out else vardata

            # drop curves containing zero values
            if reject_zeros:
                rows_bad = np.where(np.any(vardata == 0, axis=1))[0]
                if len(rows_bad) > 0:
                    logger.debug('%s: dropping %d curves with zero output' %
                                 (var, len(rows_bad)))
                    vardata = np.delete(vardata, rows_bad, axis=0)

            n_ok = vardata.shape[0]
            stddata[var] = vardata.std(axis=0) if n_ok > 0 else None
            avgdata[var] = vardata.mean(axis=0) if n_ok > 0 else None

            logger.debug('%s: averaged %d/%d curves' % (var, n_ok, Ntot))
            N_ok[var] = n_ok

    if not avgdata:
        logger.warning('nothing averaged')
    return (avgdata, stddata, N_ok, Ncyc)


def _outlier_rows(A, max_dist):
    """ Find outlier rows from A, defined as max distance from median row """
    med = np.median(A, axis=0)
    return (np.abs(A-med)).max(axis=1) > max_dist


def _collect_model_data(trials, fp_cycles_only=False):
    """ Collect given model data across trials and cycles.
    Returns a dict of numpy arrays keyed by variable.

    trials: list
        filename, or list of filenames (c3d) to read trials from, or list
        of Trial instances
    fp_cycles_only: bool
        If True, only collect data from forceplate cycles. Kinetics will always
        be collected from forceplate cycles only.
    """

    if not trials:
        logger.debug('no trials')
        return
    if not isinstance(trials, list):
        trials = [trials]

    data_all = dict()

    for model in models.models_all:
        for var in model.varnames:
            data_all[var] = None

    nc = dict()
    nc['R'], nc['L'], nc['Rkin'], nc['Lkin'] = (0,)*4

    for n, trial_ in enumerate(trials):
        # see whether it's already a Trial instance or we need to create one
        if isinstance(trial_, Trial):
            trial = trial_
        else:
            try:
                trial = Trial(trial_)
            except GaitDataError:
                logger.warning('cannot load %s' % trial_)

        logger.debug('collecting data for %s' % trial.trialname)

        # see which models are included in trial
        models_ok = list()
        for model in models.models_all:
            var = model.varnames[0]
            try:
                data = trial.get_model_data(var)[1]
                models_ok.append(model)
            except GaitDataError:
                logger.debug('cannot read variable %s from %s, skipping '
                             'corresponding model %s' % (var, trial.trialname,
                                                         model.desc))
        for model in models_ok:
            # gather data
            for cycle in trial.cycles:
                trial.set_norm_cycle(cycle)
                side = cycle.context
                if cycle.on_forceplate:
                    nc[side+'kin'] += 1
                nc[side] += 1

                for var in model.varnames:
                    # pick data only if var context matches cycle context
                    # FIXME: this may not work with all models
                    if var[0] == side:
                        # don't collect kinetics if cycle is not on forceplate
                        if ((model.is_kinetic_var(var) or fp_cycles_only) and
                           not cycle.on_forceplate):
                                continue
                        data = trial.get_model_data(var)[1]
                        # add as first row or concatenate to existing data
                        data_all[var] = (data[None, :] if data_all[var]
                                         is None else
                                         np.concatenate([data_all[var],
                                                        data[None, :]]))

    n = len(trials)
    logger.debug('collected %d trials, %d/%d R/L cycles, %d/%d kinetics cycles'
                 % (n, nc['R'], nc['L'], nc['Rkin'], nc['Lkin']))

    return data_all, nc



# -*- coding: utf-8 -*-
"""
Created on Fri Sep 23 11:17:10 2016

Misc numerical utils

@author: Jussi (jnu@iki.fi)
"""

from __future__ import division

import datetime
import numpy as np
from scipy.linalg import norm
from scipy.signal import medfilt
from numpy.lib.stride_tricks import as_strided


def check_hetu(hetu):
    """ This checks validity of a Finnish social security number (hetu) """
    chrs = "0123456789ABCDEFHJKLMNPRSTUVWXY"
    if len(hetu) != 11:
        return False
    # check day and month
    pp, kk = int(hetu[:2]), int(hetu[2:4])
    if not (0 <= pp <= 31 and 1 <= kk <= 12):
        return False
    # check 'checksum'
    chk = chrs[(int(hetu[:6] + hetu[7:10])) % 31]
    if hetu[-1] != chk:
        return False
    return True


def age_from_hetu(hetu, d1=None):
    """ Return age at date d1 (datetime.date object) from hetu. If d1 is None,
    it is taken from current system time. """
    if not hetu:
        return None
    if not check_hetu(hetu):
        raise ValueError('Invalid hetu')
    if d1 is None:
        d1 = datetime.date.today()
    day, month, yr = int(hetu[:2]), int(hetu[2:4]), int(hetu[4:6])
    yr += {'+': 1800, '-': 1900, 'A': 2000}[hetu[6]]
    d0 = datetime.date(yr, month, day)
    return d1.year - d0.year - ((d1.month, d1.day) < (d0.month, d0.day))


def rolling_fun_strided(m, fun, win, axis=None):
    """ Window array along given axis and apply fun() to the windowed data.
    No padding, i.e. returned array is shorter in the axis dim by (win-1) """
    if axis is None:
        m = m.flatten()
        axis = 0
    sh = m.shape
    st = m.strides
    # break up the given dim into windows, insert a new dim
    sh_ = sh[:axis] + (sh[axis] - win + 1, win) + sh[axis+1:]
    # insert a stride for the new dim, same as for the given dim
    st_ = st[:axis] + (st[axis], st[axis]) + st[axis+1:]
    # apply fun on the new dimension
    return fun(as_strided(m, sh_, st_), axis=axis+1)


def rising_zerocross(x):
    """ Return indices of rising zero crossings in sequence,
    i.e. n where x[n] >= 0 and x[n-1] < 0 """
    x = np.array(x)  # this should not hurt
    return np.where(np.logical_and(x[1:] >= 0, x[:-1] < 0))[0] + 1


def falling_zerocross(x):
    return rising_zerocross(-x)


def _padded_shift(x, n):
    """Shift x right by n samples (or left if negative) and zero pad so
    that original length is kept"""
    pads = (n, 0) if n > 0 else (0, -n)
    x_ = np.pad(x, pads, mode='constant')
    return x_[:-n] if n > 0 else x_[-n:]


def best_match(v, b):
    """ Replace elements of v using their closest matches in b """
    v = np.array(v)
    b = np.array(b)
    if b.size == 0:
        return v
    inds = np.abs(v[np.newaxis, :] - b[:, np.newaxis]).argmin(axis=0)
    return b[inds]


def isfloat(x):
    """ Return True for float-conversible values, False otherwise """
    try:
        float(x)
        return True
    except ValueError:
        return False


def isint(x):
    """ Return True for int-conversible values, False otherwise """
    try:
        int(x)
        return True
    except ValueError:
        return False


def _baseline(v):
    """ Baseline v using histogram. Subtracts the most prominent
    signal level """
    v = v.squeeze()
    if len(v.shape) != 1:
        raise ValueError('Need 1-dim input')
    v = np.array(v)
    nbins = int(len(v) / 10)  # exact n of bins should not matter
    ns, edges = np.histogram(v, bins=nbins)
    peak_ind = np.where(ns == np.max(ns))[0][0]
    return v - np.mean(edges[peak_ind:peak_ind+2])


def center_of_pressure(F, M, dz):
    """ Compute CoP according to AMTI instructions. The results differ
    slightly (few mm) from Nexus, for unknown reasons (different filter?)
    See http://health.uottawa.ca/biomech/courses/apa6903/amticalc.pdf """
    FP_FILTFUN = medfilt  # filter function
    FP_FILTW = 5  # median filter width
    fx, fy, fz = tuple(F.T)  # split columns into separate vars
    mx, my, mz = tuple(M.T)
    fz = FP_FILTFUN(fz, FP_FILTW)
    nz_inds = np.where(np.abs(fz) > 0)[0]  # only divide on nonzero inds
    cop = np.zeros((fx.shape[0], 3))
    cop[nz_inds, 0] = -(my[nz_inds] + fx[nz_inds] * dz)/fz[nz_inds]
    cop[nz_inds, 1] = (mx[nz_inds] - fy[nz_inds] * dz)/fz[nz_inds]
    return cop


def change_coords(pts, wR, wT):
    """ Translate pts (N x 3) into a new coordinate system described by
    rotation matrix wR and translation vector wT """
    pts = np.array(pts)
    return np.dot(wR, pts.T).T + wT


def segment_angles(P):
    """ Compute angles between segments defined by ordered points in P
    (N x 3 array). Can also be 3-d matrix of T x N x 3 to get time-dependent
    data. Output will be (N-2) vector or T x (N-2) matrix of angles in radians.
    If successive points are identical, nan:s will be output for the
    corresponding angles.
    """
    if P.shape[-1] != 3 or len(P.shape) not in [2, 3]:
        raise ValueError('Invalid shape of input matrix')
    if len(P.shape) == 2:
        P = P[np.newaxis, ...]  # insert singleton time axis
    Pd = np.diff(P, axis=1)  # point-to-point vectors
    vnorms = np.linalg.norm(Pd, axis=2)[..., np.newaxis]
    # ignore 0/0 and x/0 errors -> nan
    with np.errstate(divide='ignore', invalid='ignore'):
        Pdn = Pd / vnorms
    # take dot products between successive vectors and angles by arccos
    dots = np.sum(Pdn[:, 0:-1, :] * Pdn[:, 1:, :], axis=2)
    dots = dots[0, :] if dots.shape[0] == 1 else dots  # rm singleton dim
    return np.pi - np.arccos(dots)


def running_sum(M, win, axis=None):
    """ Running (windowed) sum of sequence M using cumulative sum,
        along given axis. Inspired by
        http://arogozhnikov.github.io/2015/09/30/NumpyTipsAndTricks2.html """
    if axis is None:
        M = M.flatten()
    s = np.cumsum(M, axis=axis)
    s = np.insert(s, 0, [0], axis=axis)
    len_ = s.shape[0] if axis is None else s.shape[axis]
    return (s.take(np.arange(win, len_), axis=axis) -
            s.take(np.arange(0, len_-win), axis=axis))


def rms(data, win):
    """ Return RMS for a given data (1-d; will be flattened if not) """
    if win % 2 != 1:
        raise ValueError('Need RMS window of odd length')
    rms_ = np.sqrt(running_sum(data**2, win) / win)
    # pad ends of RMS data so that lengths are matched
    pad = np.repeat(0, (win-1)/2)
    return np.concatenate([pad, rms_, pad])



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

Run automark for single Nexus trial. Events are restricted to ROI.

Note: event detection is less accurate compared to running autoprocess for the
whole session, since here we cannot use statistics for velocity thresholding.

@author: Jussi (jnu@iki.fi)
"""

from __future__ import print_function
import argparse
import logging

from gaitutils import nexus, utils, read_data, cfg


def automark_single(plot=False):

    vicon = nexus.viconnexus()
    roi = vicon.GetTrialRegionOfInterest()
    vicon.ClearAllEvents()

    foot_markers = (cfg.autoproc.left_foot_markers +
                    cfg.autoproc.right_foot_markers)
    mkrs = foot_markers + utils._pig_pelvis_markers()
    mkrdata = read_data.get_marker_data(vicon, mkrs, ignore_missing=True)
    fpe = utils.detect_forceplate_events(vicon, mkrdata, roi=roi)
    vel = utils.get_foot_contact_velocity(mkrdata, fpe, roi=roi)
    utils.automark_events(vicon, vel_thresholds=vel, fp_events=fpe, roi=roi,
                          plot=plot)


if __name__ == '__main__':
    logging.basicConfig(level=logging.DEBUG)

    parser = argparse.ArgumentParser()
    parser.add_argument('--plot', action='store_true',
                        help='plot velocity curves')
    args = parser.parse_args()
    automark_single(args.plot)



# -*- coding: utf-8 -*-
"""

Wrapper methods that read from either Vicon Nexus or c3d files.
'source' argument can be either a ViconNexus.ViconNexus instance or
a path to a c3d file.


@author: Jussi (jnu@iki.fi)

"""

from . import nexus
from . import c3d


def _reader_module(source):
    """ Determine the appropriate module to use """
    if nexus.is_vicon_instance(source):
        return nexus
    elif c3d.is_c3dfile(source):
        return c3d
    else:
        raise ValueError('Error reading from source: %s' % source)


def get_metadata(source):
    """ Get trial and subject info. Returns dict with:
    trialname: name of trial
    sessionpath: full path to directory where session is contained
    length: trial length in frames
    offset: frame offset of returned data from beginning of trial
    framerate: capture framerate
    analograte: sampling rate for analog devices
    name: subject name
    subj_params: subject parameters (bodymass etc.)
    lstrikes, rstrikes: foot strike events l/r
    ltoeoffs, rtoeoffs: toeoffs l/r
    """
    meta = _reader_module(source).get_metadata(source)
    # Nexus uses slightly different metadata field names as c3d,
    # so translate here into c3d naming convention
    # XXX: c3d angle params are apparently in radians while Nexus uses degrees
    # - NOT translated here!
    if nexus.is_vicon_instance(source):
        pars = meta['subj_params']
        for par in pars.keys():
            if 'Right' in par:
                new_par = par.replace('Right', 'R')
                pars[new_par] = pars.pop(par)
            if 'Left' in par:
                new_par = par.replace('Left', 'L')
                pars[new_par] = pars.pop(par)
    return meta


def get_forceplate_data(source):
    """ Get forceplate data. Returns dict with:
    F: Nx3 array with force x,y,z components
    M: Nx3 array with moment x,y,z components
    Ftot: Nx1 array of total force
    CoP:  Nx3 array, center of pressure
    analograte: sampling rate
    samplesperframe: samples per capture frame
    """
    return _reader_module(source).get_forceplate_data(source)


def get_marker_data(source, markers, ignore_edge_gaps=True,
                    ignore_missing=False):
    """ Get position, velocity and acceleration for a given marker(s)
    (str or list of str).
    Returns dict mkrdata keyed with marker names followed by _P, _V or _A
    (position, velocity, acceleration). Values are Nx3 matrices
    data, e.g. mkrdata['RHEE_V'] is a Nx3 matrix with velocity x, y and z
    components. Also computes gaps, with keys as e.g. 'RHEE_gaps'.
    markers: list of marker names
    ignore_edge_gaps: whether leading/trailing "gaps" should be ignored
    or marked as gaps. Nexus writes out gaps also at beginning/end of trial
    when reconstructions are unavailable.
    ignore_missing: whether to ignore missing markers on read or raise an
    exception.
    """
    return _reader_module(source)._get_marker_data(source, markers,
                                                   ignore_edge_gaps=ignore_edge_gaps,
                                                   ignore_missing=ignore_missing)


def get_emg_data(source):
    """ Get EMG data. Returns dict with keys """
    return _reader_module(source).get_emg_data(source)


def get_analysis(source, condition='unknown'):
    if nexus.is_vicon_instance(source):
        raise Exception('Analysis var reads from Nexus not supported yet')
    return _reader_module(source).get_analysis(source, condition)


def get_accelerometer_data(source):
    """ Get accelerometer data. Returns dict with keys """
    return _reader_module(source).get_accelerometer_data(source)


def get_model_data(source, model):
    """ Get other variables such as model outputs """
    modeldata = _reader_module(source).get_model_data(source, model)
    for var in model.read_vars:
            # convert Moment variables into SI units
            if var.find('Moment') > 0:
                modeldata[var] /= 1.0e3  # Nmm -> Nm
            # split 3-d arrays into x,y,z variables
            if model.read_strategy == 'split_xyz':
                if modeldata[var].shape[0] == 3:
                    modeldata[var+'X'] = modeldata[var][0, :]
                    modeldata[var+'Y'] = modeldata[var][1, :]
                    modeldata[var+'Z'] = modeldata[var][2, :]
                else:
                    raise ValueError('Expected a 3D array')
    return modeldata



# -*- coding: utf-8 -*-
"""

Definitions for various models (PiG, muscle length, etc.)

To create a new model, create a GaitModel() instance, fill in the data and
append to models_all.

@author: Jussi (jnu@iki.fi)
"""

from past.builtins import basestring
from builtins import object
from collections import defaultdict


models_all = []


def model_from_var(var_):
    """ Return model corresponding to specified variable.
    Returns the GaitModel instance that has the specified variable. """
    if var_ is None:
        return None
    elif not isinstance(var_, basestring):
        raise ValueError('Variable name must be a string or None')
    for model in models_all:
        if var_ in model.varnames or var_ in model.varnames_noside:
            return model
    return None


# convenience methods for model creation
def _list_with_side(vars_):
    """ Prepend variables in vars with 'L' and 'R', creating a new list of
    variables. Many model variables share the same name, except for leading
    'L' or 'R' that indicates side. """
    return [side+var for var in vars_ for side in 'LR']


def _dict_with_side_gen(di, append_side=False):
    for key, val in di.items():
        for ctxt in 'RL':
            if append_side:
                yield ctxt+key, '%s (%s)' % (val, ctxt)
            else:
                yield ctxt+key, val


def _dict_with_side(di, append_side=False):
    """Helper: prepend all dict keys with 'R' and 'L' and return new dict.
    If append_side, also append corresponding ' (R)' or ' (L)' to every dict
    value."""
    return dict(_dict_with_side_gen(di, append_side=append_side))


class GaitModel(object):
    """ A class (template) for storing information about a model, e.g.
    Plug-in Gait. The data describes model-specific variable names etc.
    and is intended (currently not forced) to be non-mutable.
    The actual data is stored elsewhere. """

    def __init__(self):
        self.read_vars = list()  # vars to be read from data
        # How to read multidimensional variables: 'split_xyz' splits each
        # variable into x,y,z components and names them appropriately.
        # 'last' reads the last component only (c3d scalars are saved
        # as last component of 3-d array (??))
        self.read_strategy = None
        self.desc = ''  # description of model
        self.varnames = list()   # resulting variable names
        self.varnames_noside = list()  # variables without side
        self.varlabels_noside = dict()  # variables without side
        self.varlabels = dict()  # descriptive label for each variable
        # mapping from variable names to gcd normal data variables
        self.gcd_normaldata_map = dict()
        # units for the variables
        self.units = dict()
        # description of neg/pos dirs for each variable in
        # tuples, e.g. ('Plantarflexion', 'Dorsiflexion')
        self.ydesc = dict()
        # callable to return whether var requires valid forceplate contact
        self.is_kinetic_var = lambda var: False
        # callable to return whether var read is allowed to fail
        self.is_optional_var = lambda var: False


""" Create models """

#
# Oxford foot model
#
ofm = GaitModel()
ofm.desc = 'Oxford foot model kinematics'
ofm.type = 'OFM'
ofm.read_strategy = 'split_xyz'

ofm.read_vars = _list_with_side(['FFHFA',
                                 'FFTBA',
                                 'HFTBA',
                                 'HFTFL',
                                 'HXFFA',
                                 'TIBA'])

ofm.varlabels_noside = {'FFHFAX': 'Forefoot-hindfoot dorsiflexion',
                        'FFHFAY': 'Forefoot-hindfoot adduction',
                        'FFHFAZ': 'Forefoot-hindfoot supination',
                        'FFTBAX': 'Forefoot-tibia dorsiflexion',
                        'FFTBAY': 'Forefoot-tibia adduction',
                        'FFTBAZ': 'Forefoot-tibia supination',
                        'HFTBAX': 'Hindfoot-tibia dorsiflexion',
                        'HFTBAY': 'Hindfoot-tibia rotation',
                        'HFTBAZ': 'Hindfoot-tibia inversion',
                        'HFTFLX': 'Hindfoot-lab dorsiflexion',
                        'HFTFLY': 'Hindfoot-lab rotation',
                        'HFTFLZ': 'Hindfoot-lab inversion',
                        'HXFFAX': 'Hallux-forefoot dorsiflexion',
                        'HXFFAY': 'Hallux-forefoot varus',
                        'HXFFAZ': 'NA',
                        'TIBAX': 'Tibia-lab forward tilt',
                        'TIBAY': 'Tibia-lab lateral tilt',
                        'TIBAZ': 'Tibia internal rotation'}

ofm.units = defaultdict(lambda: 'deg')
ofm.ydesc = _dict_with_side({'FFHFAX': ('Plantarflexion', 'Dorsiflexion'),
                             'FFHFAY': ('Abduction', 'Adduction'),
                             'FFHFAZ': ('Pronation', 'Supination'),
                             'FFTBAX': ('Plantarflexion', 'Dorsiflexion'),
                             'FFTBAY': ('Abduction', 'Adduction'),
                             'FFTBAZ': ('Pronation', 'Supination'),
                             'HFTBAX': ('Plantarflexion', 'Dorsiflexion'),
                             'HFTBAY': ('External', 'Internal'),
                             'HFTBAZ': ('Eversion', 'Inversion'),
                             'HFTFLX': ('Plantarflexion', 'Dorsiflexion'),
                             'HFTFLY': ('External', 'Internal'),
                             'HFTFLZ': ('Eversion', 'Inversion'),
                             'HXFFAX': ('Plantarflexion', 'Dorsiflexion'),
                             'HXFFAY': ('Valgus', 'Varus'),
                             'HXFFAZ': ('NA', 'NA'),
                             'TIBAX': ('Backward', 'Forward'),
                             'TIBAY': ('Distal', 'Lateral'),
                             'TIBAZ': ('External', 'Internal')})


# OFM may be unilateral, so make all vars optional
ofm.is_optional_var = lambda var: True
ofm.varlabels = _dict_with_side(ofm.varlabels_noside)
ofm.varnames = ofm.varlabels.keys()
ofm.varnames_noside = ofm.varlabels_noside.keys()

models_all.append(ofm)


#
# Plug-in Gait upper body kinematics
#
pig_upperbody = GaitModel()
pig_upperbody.desc = 'Plug-in Gait upper body kinematics'
pig_upperbody.type = 'PiG'
pig_upperbody.read_strategy = 'split_xyz'

pig_upperbody.read_vars = _list_with_side(['ThoraxAngles',
                                           'ShoulderAngles',
                                           'SpineAngles',
                                           'WristAngles',
                                           'ElbowAngles',
                                           'NeckAngles',
                                           'HeadAngles'])

pig_upperbody.varlabels_noside = {'ThoraxAnglesX': 'Thorax flex/ext',
                                  'ThoraxAnglesY': 'Thorax lateral flex',
                                  'ThoraxAnglesZ': 'Thorax rotation',
                                  'ShoulderAnglesX': 'Shoulder flex/ext',
                                  'ShoulderAnglesY': 'Shoulder abd/add',
                                  'ShoulderAnglesZ': 'Shoulder rotation',
                                  'SpineAnglesX': 'Spine flex/ext',
                                  'SpineAnglesY': 'Spine lateral flex',
                                  'SpineAnglesZ': 'Spine rotation',
                                  'WristAnglesX': 'Wrist flex/ext',
                                  'WristAnglesY': 'Wrist ulnar/radial',
                                  'WristAnglesZ': 'Wrist sup/pron',
                                  'ElbowAnglesX': 'Elbow flex/ext',
                                  'ElbowAnglesY': 'Elbow y angle',
                                  'ElbowAnglesZ': 'Elbow rotation',
                                  'NeckAnglesX': 'Neck sagittal',
                                  'NeckAnglesY': 'Neck frontal',
                                  'NeckAnglesZ': 'Neck rotation',
                                  'HeadAnglesX': 'Head sagittal',
                                  'HeadAnglesY': 'Head frontal',
                                  'HeadAnglesZ': 'Head rotation'}

# FIXME: for now, ylabel is just the degree sign for all vars
pig_upperbody.units = defaultdict(lambda: 'deg')
pig_upperbody.ydesc = defaultdict(lambda: ('', ''))  # FIXME: see Vicon docs

pig_upperbody.varlabels = _dict_with_side(pig_upperbody.varlabels_noside)
pig_upperbody.varnames = pig_upperbody.varlabels.keys()
pig_upperbody.varnames_noside = pig_upperbody.varlabels_noside.keys()

models_all.append(pig_upperbody)


#
# Plug-in Gait lower body kinematics
#
pig_lowerbody = GaitModel()
pig_lowerbody.desc = 'Plug-in Gait lower body kinematics'
pig_lowerbody.type = 'PiG'
pig_lowerbody.read_strategy = 'split_xyz'
pig_lowerbody.read_vars = _list_with_side(['HipAngles',
                                           'KneeAngles',
                                           'AnkleAngles',
                                           'ForeFootAngles',
                                           'PelvisAngles',
                                           'FootProgressAngles'])

pig_lowerbody.varlabels_noside = {
                             'AnkleAnglesX': 'Ankle dorsi/plant',
                             'AnkleAnglesY': 'Ankle adduction',
                             'AnkleAnglesZ': 'Ankle rotation',
                             'FootProgressAnglesZ': 'Foot progress angles',
                             'ForeFootAnglesX': 'Fore/hindfoot dorsi/plant',
                             'ForeFootAnglesY': 'Fore/hindfoot adduction',
                             'ForeFootAnglesZ': 'Fore/hindfoot inversion',
                             'HipAnglesX': 'Hip flexion',
                             'HipAnglesY': 'Hip adduction',
                             'HipAnglesZ': 'Hip rotation',
                             'KneeAnglesX': 'Knee flexion',
                             'KneeAnglesY': 'Knee adduction',
                             'KneeAnglesZ': 'Knee rotation',
                             'PelvisAnglesX': 'Pelvic tilt',
                             'PelvisAnglesY': 'Pelvic obliquity',
                             'PelvisAnglesZ': 'Pelvic rotation'}

pig_lowerbody.varlabels = _dict_with_side(pig_lowerbody.varlabels_noside)

pig_lowerbody.varnames = pig_lowerbody.varlabels.keys()
pig_lowerbody.varnames_noside = pig_lowerbody.varlabels_noside.keys()
# for CGM2 compatibility
pig_lowerbody.is_optional_var = lambda var: 'ForeFootAngles' in var


pig_lowerbody.gcd_normaldata_map = {
            'DorsiPlanFlex': 'AnkleAnglesX',
            'FootProgression': 'FootProgressAnglesZ',
            'FootRotation': 'AnkleAnglesZ',
            'HipAbAdduct': 'HipAnglesY',
            'HipFlexExt': 'HipAnglesX',
            'HipRotation': 'HipAnglesZ',
            'KneeFlexExt': 'KneeAnglesX',
            'KneeRotation': 'KneeAnglesZ',
            'KneeValgVar': 'KneeAnglesY',
            'PelvicObliquity': 'PelvisAnglesY',
            'PelvicRotation': 'PelvisAnglesZ',
            'PelvicTilt': 'PelvisAnglesX'}

# unit is degrees for each kinematic variable
pig_lowerbody.units = defaultdict(lambda: 'deg')

pig_lowerbody.ydesc = _dict_with_side({
                         'AnkleAnglesX': ('Plantarflexion', 'Dorsiflexion'),
                         'AnkleAnglesY': ('Abduction', 'Adduction'),
                         'AnkleAnglesZ': ('External', 'Internal'),
                         'FootProgressAnglesZ': ('External', 'Internal'),
                         'ForeFootAnglesX': ('Plantarflexion', 'Dorsiflexion'),
                         'ForeFootAnglesY': ('Abduction', 'Adduction'),
                         'ForeFootAnglesZ': ('Eversion', 'Inversion'),
                         'HipAnglesX': ('Extension', 'Flexion'),
                         'HipAnglesY': ('Abduction', 'Adduction'),
                         'HipAnglesZ': ('External', 'Internal'),
                         'KneeAnglesX': ('Extension', 'Flexion'),
                         'KneeAnglesY': ('Valgus', 'Varus'),
                         'KneeAnglesZ': ('External', 'Internal'),
                         'PelvisAnglesX': ('Posterior', 'Anterior'),
                         'PelvisAnglesY': ('Down', 'Up'),
                         'PelvisAnglesZ': ('External', 'Internal')})

models_all.append(pig_lowerbody)


#
# Plug-in Gait lowerbody kinetics
#
pig_lowerbody_kinetics = GaitModel()
pig_lowerbody_kinetics.desc = 'Plug-in Gait lower body kinetics'
pig_lowerbody_kinetics.type = 'PiG'
pig_lowerbody_kinetics.read_strategy = 'split_xyz'
pig_lowerbody_kinetics.read_vars = _list_with_side(['HipMoment',
                                                    'KneeMoment',
                                                    'AnkleMoment',
                                                    'HipPower',
                                                    'KneePower',
                                                    'AnklePower',
                                                    'NormalisedGRF'])

pig_lowerbody_kinetics.varlabels_noside = {
                             'AnkleMomentX': 'Ankle dors/plan moment',
                             'AnkleMomentY': 'Ankle ab/add moment',
                             'AnkleMomentZ': 'Ankle rotation moment',
                             'AnklePowerZ': 'Ankle power',
                             'HipMomentX': 'Hip flex/ext moment',
                             'HipMomentY': 'Hip ab/add moment',
                             'HipMomentZ': 'Hip rotation moment',
                             'HipPowerZ': 'Hip power',
                             'KneeMomentX': 'Knee flex/ext moment',
                             'KneeMomentY': 'Knee ab/add moment',
                             'KneeMomentZ': 'Knee rotation moment',
                             'KneePowerZ': 'Knee power',
                             'NormalisedGRFX': 'Norm. GRF (x)',
                             'NormalisedGRFY': 'Norm. GRF (y)',
                             'NormalisedGRFZ': 'Norm. GRF (z)'}

pig_lowerbody_kinetics.varlabels = _dict_with_side(pig_lowerbody_kinetics.
                                                   varlabels_noside)

pig_lowerbody_kinetics.varnames = pig_lowerbody_kinetics.varlabels.keys()
pig_lowerbody_kinetics.varnames_noside = (pig_lowerbody_kinetics.
                                          varlabels_noside.keys())

pig_lowerbody_kinetics.gcd_normaldata_map = {
            'AnklePower': 'AnklePowerZ',
            'DorsiPlanFlexMoment': 'AnkleMomentX',
            'FootAbAdductMoment': 'AnkleMomentY',
            'FootRotationMoment': 'AnkleMomentZ',
            'HipAbAdductMoment': 'HipMomentY',
            'HipFlexExtMoment': 'HipMomentX',
            'HipPower': 'HipPowerZ',
            'HipRotationMoment': 'HipMomentZ',
            'KneeFlexExtMoment': 'KneeMomentX',
            'KneePower': 'KneePowerZ',
            'KneeRotationMoment': 'KneeMomentZ',
            'KneeValgVarMoment': 'KneeMomentY'}

pig_lowerbody_kinetics.units = _dict_with_side({
                         'AnkleMomentX': 'Nm/kg',
                         'AnkleMomentY': 'Nm/kg',
                         'AnkleMomentZ': 'Nm/kg',
                         'AnklePowerZ': 'W/kg',
                         'HipMomentX': 'Nm/kg',
                         'HipMomentY': 'Nm/kg',
                         'HipMomentZ': 'Nm/kg',
                         'HipPowerZ': 'W/kg',
                         'KneeMomentX': 'Nm/kg',
                         'KneeMomentY': 'Nm/kg',
                         'KneeMomentZ': 'Nm/kg',
                         'KneePowerZ': 'W/kg',
                         'NormalisedGRFX': '%BW',
                         'NormalisedGRFY': '%BW',
                         'NormalisedGRFZ': '%BW'})

pig_lowerbody_kinetics.ydesc = _dict_with_side({
                         'AnkleMomentX': ('Dorsiflexion', 'Plantarflexion'),
                         'AnkleMomentY': ('Adduction', 'Abduction'),
                         'AnkleMomentZ': ('Internal', 'External'),  # FIXME: check
                         'AnklePowerZ': ('Absorbing', 'Generating'),
                         'HipMomentX': ('Flexion', 'Extension'),
                         'HipMomentY': ('Adduction', 'Abduction'),
                         'HipMomentZ': ('Internal', 'External'),
                         'HipPowerZ': ('Absorbing', 'Generating'),
                         'KneeMomentX': ('Flexion', 'Extension'),
                         'KneeMomentY': ('Varus', 'Valgus'),
                         'KneeMomentZ': ('Internal', 'External'),
                         'KneePowerZ': ('Absorbing', 'Generating'),
                         'NormalisedGRFX': ('', ''),
                         'NormalisedGRFY': ('', ''),
                         'NormalisedGRFZ': ('Downward', 'Upward')})

pig_lowerbody_kinetics.is_kinetic_var = (lambda varname: True)
# GRF may not be present if no kinetics
pig_lowerbody_kinetics.is_optional_var = lambda varname: 'GRF' in varname

models_all.append(pig_lowerbody_kinetics)

#
# Muscle length (MuscleLength.mod)
#

musclelen = GaitModel()
musclelen.desc = 'Muscle length (MuscleLength.mod)'
musclelen.type = 'musclelen'
musclelen.read_strategy = 'last'

musclelen.varlabels_noside = {
                        'AdBrLength': 'AdBrLength',
                        'AdLoLength': 'AdLoLength',
                        'AdMaInfLength': 'AdMaInfLength',
                        'AdMaMidLength': 'AdMaMidLength',
                        'AdMaSupLength': 'AdMaSupLength',
                        'BiFLLength': 'Biceps femoris length',
                        'BiFSLength': 'BiFSLength',
                        'ExDLLength': 'ExDLLength',
                        'ExHLLength': 'ExHLLength',
                        'FlDLLength': 'FlDLLength',
                        'FlHLLength': 'FlHLLength',
                        'GMedAntLength': 'GMedAntLength',
                        'GMedMidLength': 'GMedMidLength',
                        'GMedPosLength': 'GMedPosLength',
                        'GMinAntLength': 'GMinAntLength',
                        'GMinMidLength': 'GMinMidLength',
                        'GMinPosLength': 'GMinPosLength',
                        'GemeLength': 'GemeLength',
                        'GlMaInfLength': 'GlMaInfLength',
                        'GlMaMidLength': 'GlMaMidLength',
                        'GlMaSupLength': 'GlMaSupLength',
                        'GracLength': 'Gracilis length',
                        'IliaLength': 'IliaLength',
                        'LaGaLength': 'Lateral gastrocnemius length',
                        'MeGaLength': 'Medial gastrocnemius length',
                        'PELOLength': 'PELOLength',
                        'PeBrLength': 'PeBrLength',
                        'PeTeLength': 'PeTeLength',
                        'PectLength': 'PectLength',
                        'PeriLength': 'PeriLength',
                        'PsoaLength': 'Psoas length',
                        'QuFeLength': 'QuFeLength',
                        'ReFeLength': 'Rectus femoris length',
                        'SartLength': 'SartLength',
                        'SeMeLength': 'Semimembranosus length',
                        'SeTeLength': 'Semitendinosus length',
                        'SoleLength': 'Soleus length',
                        'TiAnLength': 'Tibialis anterior length',
                        'TiPoLength': 'TiPoLength',
                        'VaInLength': 'VaInLength',
                        'VaLaLength': 'VaLaLength',
                        'VaMeLength': 'VaMeLength'}

musclelen.varlabels = _dict_with_side(musclelen.varlabels_noside)
musclelen.read_vars = musclelen.varlabels.keys()
musclelen.varnames = musclelen.read_vars
musclelen.varnames_noside = musclelen.varlabels_noside.keys()

musclelen.units = defaultdict(lambda: 'Norm length')  # FIXME: % of what?
musclelen.ydesc = defaultdict(lambda: ('', ''))

models_all.append(musclelen)



# -*- coding: utf-8 -*-
"""
plotly plotting functions

@author: Jussi (jnu@iki.fi)
"""

from builtins import range
import plotly
import plotly.graph_objs as go
import plotly.tools
import numpy as np
from itertools import cycle
import datetime
import logging

from gaitutils import models, cfg, normaldata
from gaitutils.envutils import GaitDataError
from gaitutils.trial import Gaitcycle, Noncycle


logger = logging.getLogger(__name__)


def _plotly_fill_between(x, ylow, yhigh, **kwargs):
    """Fill area between ylow and yhigh"""
    x_ = np.concatenate([x, x[::-1]])  # construct a closed curve
    y_ = np.concatenate([yhigh, ylow[::-1]])
    return go.Scatter(x=x_, y=y_, fill='toself', hoverinfo='none', **kwargs)


def _var_title(var):
    """Get proper title for variable"""
    mod = models.model_from_var(var)
    if mod:
        if var in mod.varlabels_noside:
            return mod.varlabels_noside[var]
        elif var in mod.varlabels:
            return mod.varlabels[var]
    elif var in cfg.emg.channel_labels:
        return cfg.emg.channel_labels[var]
    else:
        return ''


def _truncate_trialname(trialname):
    """Shorten trial name."""
    try:
        # try to truncate date string of the form yyyy_mm_dd
        tn_split = trialname.split('_')
        datetxt = '-'.join(tn_split[:3])
        d = datetime.datetime.strptime(datetxt, '%Y-%m-%d')
        return '%d..%s' % (d.year, '_'.join(tn_split[3:]))
    except ValueError:  # trial was not named as expected
        return trialname


def plot_trials_browser(trials, layout, **kwargs):
    """ Convenience plotter, uses plotly.offline to plot directly to browser"""
    fig = plot_trials(trials, layout, **kwargs)
    plotly.offline.plot(fig)


_plot_cache = dict()  # global for plot_trials


def plot_trials(trials, layout, model_normaldata=None, model_cycles=None,
                emg_cycles=None, legend_type='full', trial_linestyles='same',
                supplementary_data=None, maintitle=None):
    """Make a plotly plot of layout, including given trials.

    trials: list of gaitutils.Trial instances
    layout: list of lists defining plot layout (see plot.py)
    model_normaldata: dict of normal data for model variables
    legend_type: 'tag_only' for Eclipse tag, 'name_with_tag' or 'full'
    trial_linestyles: 'same' for all identical, 'trial' for trial specific
                      style, 'session' for session specific style
    supplementary_data: dict of additional data for each cycle and variable
    """
    global _plot_cache

    if not trials:
        raise GaitDataError('No trials')

    if supplementary_data is None:
        supplementary_data = dict()

    if model_normaldata is None:
        model_normaldata = normaldata.read_all_normaldata()

    # configurabe opts (here for now)
    label_fontsize = 16  # x, y labels
    subtitle_fontsize = 20  # subplot titles

    nrows = len(layout)
    if nrows == 0:
        raise GaitDataError('Empty layout')
    ncols = len(layout[0])
    if ncols == 0:
        raise GaitDataError('Empty layout')

    if len(trials) > len(plotly.colors.DEFAULT_PLOTLY_COLORS):
        logger.warning('Not enough colors for plot')
    colors = cycle(plotly.colors.DEFAULT_PLOTLY_COLORS)

    allvars = [item for row in layout for item in row]
    titles = [_var_title(var) for var in allvars]
    fig = plotly.tools.make_subplots(rows=nrows, cols=ncols, print_grid=False,
                                     subplot_titles=titles)
    tracegroups = set()
    model_normaldata_legend = True
    emg_normaldata_legend = True

    session_linestyles = dict()
    dash_styles = cycle(['solid', 'dash', 'dot', 'dashdot'])

    # these are the cycle specifications, e.g. 'forceplate'
    # the cycles
    model_cyclespec = (cfg.plot.default_model_cycles if model_cycles is None
                       else model_cycles)
    emg_cyclespec = (cfg.plot.default_emg_cycles if emg_cycles is None else
                     emg_cycles)

    trials = sorted(trials, key=lambda tr: _truncate_trialname(tr.trialname))

    for trial in trials:
        trial_color = next(colors)

        model_cycles = trial.get_cycles(model_cyclespec)
        emg_cycles = trial.get_cycles(emg_cyclespec)
        allcycles = list(set.union(set(model_cycles), set(emg_cycles)))
        if not allcycles:
            raise GaitDataError('Trial %s has no cycles of specified type' %
                                trial.trialname)

        logger.debug('plotting total of %d cycles for %s (%d model, %d EMG)'
                     % (len(allcycles), trial.trialname, len(model_cycles),
                        len(emg_cycles)))

        cycs_unnorm = [cyc.start is None for cyc in allcycles]
        is_unnormalized = any(cycs_unnorm)
        if is_unnormalized and not all(cycs_unnorm):
            raise GaitDataError('Cannot mix normalized and unnormalized data')

        for cyc_ind, cyc in enumerate(allcycles):
            #logger.debug('trial %s, cycle: %s' % (trial.trialname, cyc))
            trial.set_norm_cycle(cyc)
            context = cyc.context

            for i, row in enumerate(layout):
                for j, var in enumerate(row):
                    plot_ind = i * ncols + j + 1  # plotly subplot index
                    xaxis = 'xaxis%d' % plot_ind  # name of plotly xaxis
                    yaxis = 'yaxis%d' % plot_ind  # name of plotly yaxis

                    # in legend, traces will be grouped according to tracegroup
                    # (which is also the label)
                    if legend_type == 'name_with_tag':
                        tracegroup = '%s / %s' % (trial.trialname,
                                                  trial.eclipse_tag)
                    elif legend_type == 'short_name_with_tag':
                        tracegroup = '%s / %s' % (_truncate_trialname(trial.trialname),
                                                  trial.eclipse_tag)
                    elif legend_type == 'tag_only':
                        tracegroup = trial.eclipse_tag
                    elif legend_type == 'tag_with_cycle':
                        tracegroup = '%s / %s' % (trial.eclipse_tag,
                                                  cyc.name)
                    elif legend_type == 'full':
                        tracegroup = '%s / %s' % (trial.name_with_description,
                                                  cyc.name)
                    elif legend_type == 'short_name_with_cyclename':
                        tracegroup = '%s / %s' % (_truncate_trialname(trial.trialname),
                                                  cyc.name)
                    else:
                        raise ValueError('Invalid legend type')
                    tracename_full = '%s (%s) / %s' % (trial.trialname,
                                                       trial.eclipse_tag,
                                                       cyc.name)
                    # plotly cannot directly handle unicode objects
                    if isinstance(tracegroup, unicode):
                        tracegroup = tracegroup.encode('utf-8')

                    # only show the legend for the first trace in the
                    # tracegroup, so we do not repeat legends
                    show_legend = tracegroup not in tracegroups

                    mod = models.model_from_var(var)
                    if mod:
                        do_plot = True  # FIXME: control flow is clumsy here

                        if cyc not in model_cycles:
                            do_plot = False

                        if var in mod.varnames_noside:
                            # var context was unspecified, so choose it
                            # according to cycle context
                            var = context + var
                        elif var[0] != context:
                            # var context was specified, and has to match cycle
                            do_plot = False

                        if mod.is_kinetic_var(var):
                            # kinetic var cycles are required to have valid
                            # forceplate data
                            if not is_unnormalized and not cyc.on_forceplate:
                                do_plot = False

                        # plot normaldata before other data so that its z order
                        # is lowest (otherwise normaldata will mask other
                        # traces on hover) and it gets the 1st legend entry
                        if (model_normaldata is not None and trial == trials[0]
                            and cyc == allcycles[0] and not is_unnormalized):
                            if var[0].upper() in ['L', 'R']:
                                nvar = var[1:]
                            if nvar in model_normaldata:
                                key = nvar
                            else:
                                key = None
                            ndata = (model_normaldata[key] if key in
                                     model_normaldata else None)
                            if ndata is not None:
                                # FIXME: hardcoded color
                                normalx = np.linspace(0, 100, ndata.shape[0])
                                ntrace = _plotly_fill_between(normalx,
                                                              ndata[:, 0],
                                                              ndata[:, 1],
                                                              fillcolor='rgba(100, 100, 100, 0.2)',
                                                              name='Norm.',
                                                              legendgroup='Norm.',
                                                              showlegend=model_normaldata_legend,
                                                              line=dict(width=0))  # no border lines
                                fig.append_trace(ntrace, i+1, j+1)
                                # add to legend only once
                                model_normaldata_legend = False
                                new_subplot = False

                        if do_plot:
                            t, y = trial.get_model_data(var)

                            if trial_linestyles == 'trial':
                                # trial specific color, left side dashed
                                line = {'color': trial_color}
                                if context == 'L':
                                    line['dash'] = 'dash'
                            elif trial_linestyles == 'same':
                                # identical color for all trials
                                line = {'color':
                                        cfg.plot.model_tracecolors[context]}
                            elif trial_linestyles == 'session':
                                # session specific line style
                                line = {'color':
                                        cfg.plot.model_tracecolors[context]}
                                if trial.sessiondir in session_linestyles:
                                    dash_style = session_linestyles[trial.
                                                                    sessiondir]
                                else:
                                    dash_style = next(dash_styles)
                                    session_linestyles[trial.sessiondir] = dash_style
                                line['dash'] = dash_style

                            # check whether trace was already created
                            if (trial in _plot_cache and cyc in
                                _plot_cache[trial] and var in
                                _plot_cache[trial][cyc]):
                                        trace = _plot_cache[trial][cyc][var]
                                        # update some of the properties
                                        trace['name'] = tracegroup
                                        trace['legendgroup'] = tracegroup
                                        trace['showlegend'] = show_legend
                            else:  # need to create trace
                                trace = go.Scatter(x=t, y=y, name=tracegroup,
                                                   text=tracename_full,
                                                   legendgroup=tracegroup,
                                                   showlegend=show_legend,
                                                   hoverlabel=dict(namelength=-1),
                                                   hoverinfo='x+y+text',
                                                   line=line)
                                # add trace to cache
                                if trial not in _plot_cache:
                                    _plot_cache[trial] = dict()
                                if cyc not in _plot_cache[trial]:
                                    _plot_cache[trial][cyc] = dict()
                                _plot_cache[trial][cyc][var] = trace

                            # add toeoff marker
                            if cyc.toeoffn is not None:
                                toeoff = int(cyc.toeoffn)
                                marker = dict(color='black',
                                              symbol='triangle-up',
                                              size=8)
                                toeoff_marker = go.Scatter(x=t[toeoff:toeoff+1],
                                                           y=y[toeoff:toeoff+1],
                                                           showlegend=False,
                                                           legendgroup=tracegroup,
                                                           hoverinfo='skip',
                                                           mode='markers',
                                                           marker=marker)
                                fig.append_trace(toeoff_marker, i+1, j+1)

                            # add trace to figure
                            fig.append_trace(trace, i+1, j+1)
                            tracegroups.add(tracegroup)

                            # add supplementary data
                            if cyc in supplementary_data:
                                supdata = supplementary_data[cyc]
                                if var in supdata:
                                    logger.debug('plotting supplementary data '
                                                 'for var %s' % var)
                                    t_sup = supdata[var]['t']
                                    data_sup = supdata[var]['data']
                                    label_sup = supdata[var]['label']
                                    strace = go.Scatter(x=t_sup, y=data_sup,
                                                        name=label_sup,
                                                        text=label_sup,
                                                        line=line,
                                                        legendgroup=tracegroup,
                                                        hoverinfo='x+y+text',
                                                        showlegend=False)
                                    fig.append_trace(strace, i+1, j+1)

                            # rm x tick labels, plot too crowded
                            fig['layout'][xaxis].update(showticklabels=False)
                            yunit = mod.units[var]
                            if yunit == 'deg':
                                # plotly supports neither unicode (bug) or
                                # latex; anyway, y axes are crowded
                                yunit = ''
                                #yunit = u'\u00B0'  # Unicode degree sign
                            ydesc = [s[:3] for s in mod.ydesc[var]]  # shorten
                            ylabel = u'%s  %s  %s' % (ydesc[0], yunit, ydesc[1])
                            fig['layout'][yaxis].update(title=ylabel, titlefont={'size': label_fontsize})
                            # less decimals on hover label
                            fig['layout'][yaxis].update(hoverformat='.2f')

                    # plot EMG variable
                    elif (trial.emg.is_channel(var) or var in
                          cfg.emg.channel_labels):
                        do_plot = True
                        # plot only if EMG channel context matches cycle ctxt
                        # FIXME: this assumes that EMG names begin with context
                        if (var[0] != context or not trial.emg.status_ok(var)
                            or cyc not in emg_cycles):
                            do_plot = False
                            # FIXME: maybe annotate disconnected chans
                            # _no_ticks_or_labels(ax)
                            # _axis_annotate(ax, 'disconnected')
                        if do_plot:
                            t_, y = trial.get_emg_data(var)
                            t = (t_ / trial.samplesperframe if is_unnormalized
                                 else t_)
                            line = {'width': 1, 'color': trial_color}
                            if (trial in _plot_cache and cyc in
                                _plot_cache[trial] and var in
                                _plot_cache[trial][cyc]):
                                        #logger.debug('cache hit for: %s / %s / %s' %
                                        #             (trial.trialname, cyc.name, var))
                                        trace = _plot_cache[trial][cyc][var]
                                        trace['name'] = tracegroup
                                        trace['legendgroup'] = tracegroup
                                        trace['showlegend'] = show_legend
                            else:
                                #logger.debug('calling Scatter for: %s / %s / %s' %
                                #             (trial.trialname, cyc.name, var))
                                trace = go.Scatter(x=t,
                                                   y=y*cfg.plot.emg_multiplier,
                                                   name=tracegroup,
                                                   legendgroup=tracegroup,
                                                   showlegend=show_legend,
                                                   line=line)
                                if trial not in _plot_cache:
                                    _plot_cache[trial] = dict()
                                if cyc not in _plot_cache[trial]:
                                    _plot_cache[trial][cyc] = dict()
                                _plot_cache[trial][cyc][var] = trace

                            tracegroups.add(tracegroup)
                            fig.append_trace(trace, i+1, j+1)

                        # check whether this is the last EMG cycle and plot
                        # normals if so
                        remaining = allcycles[cyc_ind+1:]
                        last_emg = (trial == trials[-1] and not
                                    any(c in remaining for c in emg_cycles))
                        if last_emg:
                            if not is_unnormalized and var in cfg.emg.channel_normaldata:
                                emgbar_ind = cfg.emg.channel_normaldata[var]
                                for inds in emgbar_ind:
                                    # FIXME: hardcoded color
                                    # NOTE: using big values (>~1e3) for the normal bar height triggers a plotly bug
                                    # and screws up the normal bars (https://github.com/plotly/plotly.py/issues/1008)
                                    ntrace = _plotly_fill_between(inds, [-1e1]*2, [1e1]*2,  # simulate x range fill by high y values
                                                                  name='EMG norm.',
                                                                  legendgroup='EMG norm.',
                                                                  showlegend=emg_normaldata_legend,
                                                                  fillcolor='rgba(255, 0, 0, 0.1)',
                                                                  line=dict(width=0))  # no border lines                                                               
                                    fig.append_trace(ntrace, i+1, j+1)
                                    emg_normaldata_legend = False  # add to legend only once
                        
                            emg_yrange = np.array([-cfg.plot.emg_yscale, cfg.plot.emg_yscale]) * cfg.plot.emg_multiplier
                            fig['layout'][yaxis].update(title=cfg.plot.emg_ylabel, titlefont={'size': label_fontsize},
                                                        range=emg_yrange)  # FIXME: cfg
                            # prevent changes due to legend clicks etc.
                            if not is_unnormalized:
                                fig['layout'][xaxis].update(range=[0, 100])
                            # rm x tick labels, plot too crowded
                            fig['layout'][xaxis].update(showticklabels=False)

                    elif var is None:
                        continue

                    elif 'legend' in var:  # 'legend' is for mpl plotter only
                        continue

                    else:
                        raise Exception('Unknown variable %s' % var)

    # reduce subplot title font size
    for anno in fig['layout']['annotations']:
        anno['font']['size'] = subtitle_fontsize

    # put x labels on last row only, re-enable tick labels for last row
    inds_last = range((nrows-1)*ncols, nrows*ncols)
    axes_last = ['xaxis%d' % (ind+1) for ind in inds_last]
    xtitle = 'frame' if is_unnormalized else '% of gait cycle'
    for ax in axes_last:
        fig['layout'][ax].update(title=xtitle,
                                 titlefont={'size': label_fontsize},
                                 showticklabels=True)

    margin = go.layout.Margin(l=50, r=0, b=50, t=50, pad=4)  # NOQA: 741
    layout = go.Layout(margin=margin, font={'size': label_fontsize},
                       hovermode='closest', title=maintitle)

    fig['layout'].update(layout)
    return fig



# -*- coding: utf-8 -*-
"""
Handling layouts.

@author: Jussi (jnu@iki.fi)
"""

from builtins import str
from builtins import zip
import logging

from gaitutils import cfg, GaitDataError

logger = logging.getLogger(__name__)


def rm_dead_channels(emg, layout):
    """ From EMG layout, remove rows with no valid EMG data """
    layout_ = list()
    for row in layout:
        # accept channels w/ status ok, or anything that is NOT a
        # preconfigured EMG channel
        if any([ch not in cfg.emg.channel_labels or emg.status_ok(ch) for ch in row]):
            layout_.append(row)
        else:
            logger.debug('no valid data for %s, removed row' % str(row))
    if not layout_:
        logger.warning('removed all - no valid EMG channels')
    return layout_


def rm_dead_channels_multitrial(emgs, layout):
    """From layout, drop rows that do not have good data in any of the
    EMG() instances given """
    chs_ok = None
    for i, emg in enumerate(emgs):
        # accept channels w/ status ok, or anything that is NOT a
        # preconfigured EMG channel
        chs_ok_ = [ch not in cfg.emg.channel_labels or emg.status_ok(ch) for
                   row in layout for ch in row]
        # previously OK chs propagated as ok
        chs_ok = ([a or b for a, b in zip(chs_ok, chs_ok_)] if i > 0 else
                  chs_ok_)
    if not chs_ok:
        raise GaitDataError('No acceptable channels in any of the EMGs')
    rowlen = len(layout[0])
    lout = zip(*[iter(chs_ok)]*rowlen)  # grouper recipe from itertools
    rows_ok = [any(row) for row in lout]
    layout = [row for i, row in enumerate(layout) if rows_ok[i]]
    return layout


def onesided_layout(layout, side):
    """ Add 'R' or 'L' to layout variable names """
    if side not in ['R', 'L']:
        raise ValueError('Invalid side')
    return [[(side + item) if item is not None else None for item in row]
            for row in layout]


def filter_layout(layout, key, repl):
    """ Replace layout items """
    return [[repl if (item and key in item) else item for item in row]
            for row in layout]



# -*- coding: utf-8 -*-
"""
gaitutils custom Qt widgets

@author: Jussi (jnu@iki.fi)
"""

from PyQt5 import QtCore, QtWidgets
from PyQt5.QtCore import QObject, pyqtSignal
import logging
import sys


class NiceListWidgetItem(QtWidgets.QListWidgetItem):
    """ Make list items more pythonic - otherwise would have to do horrible and
    bug-prone things like checkState() """

    def __init__(self, *args, **kwargs):
        # don't pass this arg to superclass __init__
        checkable = kwargs.pop('checkable')
        super(NiceListWidgetItem, self).__init__(*args, **kwargs)
        if checkable:
            self.setFlags(self.flags() | QtCore.Qt.ItemIsUserCheckable)

    @property
    def text(self):
        return super(NiceListWidgetItem, self).text()

    @property
    def userdata(self):
        return super(NiceListWidgetItem, self).data(QtCore.Qt.UserRole)

    @userdata.setter
    def userdata(self, _data):
        if _data is not None:
            super(NiceListWidgetItem, self).setData(QtCore.Qt.UserRole, _data)

    @property
    def checkstate(self):
        return super(NiceListWidgetItem, self).checkState()

    @checkstate.setter
    def checkstate(self, checked):
        state = QtCore.Qt.Checked if checked else QtCore.Qt.Unchecked
        super(NiceListWidgetItem, self).setCheckState(state)


class NiceListWidget(QtWidgets.QListWidget):
    """ Adds some convenience to QListWidget """

    def __init__(self, parent=None):
        super(NiceListWidget, self).__init__(parent)

    @property
    def items(self):
        """ Yield all items. NB: be careful when modifying list items in
        a loop with the generator - count() is evaluated only once, so the
        generator may return items that have already been deleted """
        for i in range(self.count()):
            yield self.item(i)

    @property
    def checked_items(self):
        """ Yield checked items """
        return (item for item in self.items if item.checkstate)

    def check_all(self):
        """ Check all items """
        for item in self.items:
            item.checkstate = True

    def check_none(self):
        """ Uncheck all items """
        for item in self.items:
            item.checkstate = False

    def add_item(self, txt, data=None, checkable=False, checked=False):
        """ Add checkable item with data. Select new item. """
        item = NiceListWidgetItem(txt, self, checkable=checkable)
        item.userdata = data
        if checkable:
            item.checkstate = checked
        self.setCurrentItem(item)

    def rm_current_item(self):
        """ Remove currently selected item """
        return self.takeItem(self.row(self.currentItem()))


class QtHandler(logging.Handler):
    """Qt logging handler"""

    def __init__(self):
        logging.Handler.__init__(self)

    def emit(self, record):
        record = self.format(record)
        if record:
            XStream.stdout().write('%s\n' % record)


class XStream(QtCore.QObject):
    """Stream for Qt logging handler"""
    _stdout = None
    _stderr = None
    messageWritten = QtCore.pyqtSignal(str)

    def flush(self):
        pass

    def fileno(self):
        return -1

    def write(self, msg):
        if not self.signalsBlocked():
            self.messageWritten.emit(str(msg))

    @staticmethod
    def stdout():
        if not XStream._stdout:
            XStream._stdout = XStream()
            sys.stdout = XStream._stdout  # also capture stdout
        return XStream._stdout

    @staticmethod
    def stderr():
        if not XStream._stderr:
            XStream._stderr = XStream()
            sys.stderr = XStream._stderr  # ... and stderr
        return XStream._stderr


class ProgressBar(QtWidgets.QProgressDialog):
    """ Qt progress bar with reasonable defaults """

    def __init__(self, title):
        super(self.__class__, self).__init__()
        self.setWindowTitle(title)
        self.setCancelButton(None)
        self.setMinimum(0)
        self.setMaximum(100)
        self.setGeometry(500, 300, 500, 100)
        self.setAutoClose(False)
        self.setAutoReset(False)
        self.show()

    def update(self, text, p):
        """ Update bar, showing text and bar at p% """
        self.setLabelText(text)
        self.setValue(p)
        # update right away in case that thread is blocked
        QtWidgets.QApplication.processEvents()


class ProgressSignals(QObject):
    """Used to emit progress signals"""
    progress = pyqtSignal(object, object)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

Average cycles over the session trials.

@author: Jussi (jnu@iki.fi)
"""

import logging

import gaitutils
from gaitutils import (cfg, nexus, layouts, GaitDataError,
                       register_gui_exception_handler, sessionutils)


def do_plot(sessionpath=None, show=True, make_pdf=True):

    figs = []
    if sessionpath is None:
        sessionpath = nexus.get_sessionpath()
    c3ds = sessionutils.get_c3ds(sessionpath, trial_type='dynamic')
    if not c3ds:
        raise GaitDataError('No dynamic trials found for current session')

    atrial = gaitutils.stats.AvgTrial(c3ds)

    pl = gaitutils.Plotter()
    pl.trial = atrial

    layout = cfg.layouts.lb_kin

    maintitle_ = '%d trial average from %s' % (atrial.nfiles, sessionpath)

    for side in ['R', 'L']:
        side_str = 'right' if side == 'R' else 'left'
        maintitle = maintitle_ + ' (%s)' % side_str
        pl.layout = layouts.onesided_layout(layout, side)
        figs.append(pl.plot_trial(split_model_vars=False,
                                  model_stddev=atrial.stddev_data,
                                  maintitle=maintitle,
                                  show=False))
        if make_pdf:
            pl.create_pdf(pdf_name='kin_average_%s.pdf' % side,
                          sessionpath=sessionpath)
    if show:
        pl.show()

    return figs


if __name__ == '__main__':
    logging.basicConfig(level=logging.DEBUG)
    register_gui_exception_handler()
    do_plot()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Autoprocess all trials in current Nexus session directory. See autoproc
section in config for options.

1st pass (all trials):
-preprocess
-get fp context + strike/toeoff velocities etc.

2nd pass:
-automark (using velocity stats from previous step)
-run models + save

-write Eclipse info

GAP HANDLING:
    If cfg.autoproc.fail_on_gaps is set, the processing will fail on ANY gaps.
    Otherwise, a ROI (region of interest) will be determined based on
    cfg.autoproc.events_range. Gaps outside the ROI will not affect processing.
    However the tracking markers (cfg.autoproc.track_markers) are used to
    determine the ROI and they may not have any gaps anywhere.

@author: Jussi (jnu@iki.fi)
"""

from builtins import zip
import os
import os.path as op
import numpy as np
import argparse
import time
import logging
import itertools

from gaitutils import (nexus, eclipse, utils, GaitDataError, sessionutils,
                       read_data, cfg)


logger = logging.getLogger(__name__)


def _do_autoproc(enffiles, update_eclipse=True):
    """ Do autoprocessing for all trials listed in enffiles (list of
    paths to .enf files).
    """

    def _save_trial():
        """Save trial in Nexus"""
        logger.debug('saving trial')
        vicon.SaveTrial(cfg.autoproc.nexus_timeout)

    def _context_desc(fpev):
        """Get Eclipse description string for given forceplate events dict"""
        s = ""
        nr = len(fpev['R_strikes'])
        if nr:
            s += '%dR' % nr
        nl = len(fpev['L_strikes'])
        if nr and nl:
            s += '/'
        if nl:
            s += '%dL' % nl
        return s or cfg.autoproc.enf_descriptions['context_none']

    def _fail(trial, reason):
        """Abort processing: mark and save trial"""
        fail_desc = (cfg.autoproc.enf_descriptions[reason] if reason in
                     cfg.autoproc.enf_descriptions else reason)
        logger.debug('preprocessing failed: %s' % fail_desc)
        trial['recon_ok'] = False
        trial['description'] = fail_desc
        _save_trial()

    def _range_to_roi(subj_pos, gait_dim, mov_range):
        """Try to determine ROI (in frames) from movement range"""
        subj_pos1 = subj_pos[:, [gait_dim]]
        # find non-gap frames where we are inside movement range
        dist_ok = np.where((subj_pos1 >= mov_range[0]) &
                           (subj_pos1 <= mov_range[1]) &
                           (subj_pos1 != 0.0))[0]
        return min(dist_ok), max(dist_ok)

    # used to store stats about foot velocity
    foot_vel = {'L_strike': np.array([]), 'R_strike': np.array([]),
                'L_toeoff': np.array([]), 'R_toeoff': np.array([])}
    # 1st pass
    logger.debug('\n1st pass - processing %d trial(s)\n' % len(enffiles))
    trials = dict()

    vicon = nexus.viconnexus()
    nexus_ver = nexus.true_ver()
    logger.debug('detected Nexus version: %g' % nexus_ver)
    # close trial to prevent 'Save trial?' dialog on first open
    if nexus_ver >= 2.8:
        logger.debug('force closing open trial')
        vicon.CloseTrial(5000)  # timeout in ms

    for enffile in enffiles:
        filepath = enffile[:enffile.find('.Trial')]  # rm .TrialXXX and .enf
        filename = os.path.split(filepath)[1]
        trial = dict()
        trials[filepath] = trial
        logger.debug('loading in Nexus: %s' % filename)
        vicon.OpenTrial(filepath, cfg.autoproc.nexus_timeout)
        try:
            nexus.get_metadata(vicon)
        except GaitDataError:
            # may indicate broken or video-only trial
            logger.warning('cannot read metadata')
            trial['recon_ok'] = False
            trial['description'] = 'skipped'
            continue

        edata = eclipse.get_eclipse_keys(enffile, return_empty=True)
        logger.debug('type: %s' % edata['TYPE'])
        logger.debug('description: %s' % edata['DESCRIPTION'])
        logger.debug('notes: %s' % edata['NOTES'])
        eclipse_str = ''

        # check whether to skip trial
        if edata['TYPE'] in cfg.autoproc.type_skip:
            logger.debug('skipping based on type: %s' % edata['TYPE'])
            trial['recon_ok'] = False
            trial['description'] = 'skipped'
            continue
        skip = [s.upper() for s in cfg.autoproc.eclipse_skip]
        if (any([s in edata['DESCRIPTION'].upper() for s in skip]) or
           any([s in edata['NOTES'].upper() for s in skip])):
            logger.debug('skipping based on description/notes')
            # run preprocessing + save even for skipped trials, to mark
            # them as processed - mostly so that Eclipse export to Polygon
            # will work
            nexus.run_pipelines(vicon, cfg.autoproc.pre_pipelines)
            _save_trial()
            trial['recon_ok'] = False
            trial['description'] = 'skipped'
            continue

        # try to run preprocessing pipelines
        nexus.run_pipelines(vicon, cfg.autoproc.pre_pipelines)

        # check trial length
        trange = vicon.GetTrialRange()
        if (trange[1] - trange[0]) < cfg.autoproc.min_trial_duration:
            _fail(trial, 'short')
            continue

        # check for valid marker data
        allmarkers = nexus._get_marker_names(vicon, trajs_only=True)
        try:
            mkrdata = read_data.get_marker_data(vicon, allmarkers,
                                                ignore_missing=True)
        except GaitDataError:
            logger.debug('get_marker_data failed')
            _fail(trial, 'label_failure')
            continue

        # fail on any gaps in trial (off by default)
        gaps_found = False
        if cfg.autoproc.fail_on_gaps:
            for marker in set(allmarkers) - set(cfg.autoproc.ignore_markers):
                gaps = mkrdata[marker + '_gaps']
                if gaps.size > 0:
                    gaps_found = True
                    break
            if gaps_found:
                _fail(trial, 'gaps')
                continue

        # check for valid Plug-in Gait set
        if not utils.is_plugingait_set(mkrdata):
            logger.warning('marker set does not correspond to Plug-in Gait')
            _fail(trial, 'label_failure')
            continue
        # check for flipped markers
        flipped = list(utils._check_markers_flipped(mkrdata))
        if flipped:
            for m1, m2 in flipped:
                logger.warning('trying to swap trajectories for %s and %s'
                               % (m1, m2))
                nexus._swap_markers(vicon, m1, m2)

        # get subject position by tracking markers
        try:
            subj_pos = utils.avg_markerdata(mkrdata,
                                            cfg.autoproc.track_markers)
        except GaitDataError:
            logger.debug('gaps in tracking markers')
            _fail(trial, 'label_failure')
            continue
        gait_dim = utils.principal_movement_direction(subj_pos)
        # our roi according to events_range
        # this is not the same as Nexus ROI, which is unset at this point
        roi = _range_to_roi(subj_pos, gait_dim, cfg.autoproc.events_range)
        logger.debug('events range corresponds to frames %d-%d' % roi)
        trial['roi'] = roi

        # check forceplate data
        fp_info = (eclipse.eclipse_fp_keys(edata) if
                   cfg.autoproc.use_eclipse_fp_info else None)
        try:
            fpev = utils.detect_forceplate_events(vicon, mkrdata,
                                                  fp_info=fp_info, roi=roi)
        except GaitDataError:
            logger.warning('cannot determine forceplate events, possibly due '
                           'to gaps')
            _fail(trial, 'gaps')
            continue
        # get foot velocity info for all events (do not reduce to median)
        try:
            vel = utils.get_foot_contact_velocity(mkrdata, fpev, medians=False,
                                                  roi=roi)
        except GaitDataError:
            logger.warning('cannot determine foot velocity, possibly due to '
                           'gaps')
            _fail(trial, 'gaps')
            continue

        # preprocessing looks ok at this stage
        trial['recon_ok'] = True
        trial['mkrdata'] = mkrdata

        eclipse_str += _context_desc(fpev)
        valid = fpev['valid']
        trial['valid'] = valid
        trial['fpev'] = fpev

        # save velocity data
        for context in valid:
            nv = np.append(foot_vel[context+'_strike'], vel[context+'_strike'])
            foot_vel[context+'_strike'] = nv
            nv = np.append(foot_vel[context+'_toeoff'], vel[context+'_toeoff'])
            foot_vel[context+'_toeoff'] = nv
        eclipse_str += ','

        # main direction in lab frame (1,2,3 for x,y,z)
        inds_ok = np.where(np.any(subj_pos, axis=1))  # ignore gaps
        subj_pos_ = subj_pos[inds_ok]
        # +1/-1 for forward/backward (coord increase / decrease)
        gait_dir = np.median(np.diff(subj_pos_, axis=0), axis=0)[gait_dim]
        # write Eclipse key for direction
        if ('dir_forward' in cfg.autoproc.enf_descriptions and 'dir_backward'
           in cfg.autoproc.enf_descriptions):
            dir_str = 'dir_forward' if gait_dir > 0 else 'dir_backward'
            dir_desc = cfg.autoproc.enf_descriptions[dir_str]
            eclipse_str += '%s,' % dir_desc

        # compute gait velocity
        median_vel = utils._trial_median_velocity(vicon)
        logger.debug('median forward velocity: %.2f m/s' % median_vel)
        eclipse_str += '%.2f m/s' % median_vel

        _save_trial()
        trial['description'] = eclipse_str

        # write Eclipse fp values according to our detection
        fp_info = fpev['our_fp_info']
        if cfg.autoproc.write_eclipse_fp_info is True:
            logger.debug('writing detected forceplate info into Eclipse')
            eclipse.set_eclipse_keys(enffile, fp_info,
                                     update_existing=True)
        elif cfg.autoproc.write_eclipse_fp_info == 'reset':
            logger.debug('resetting Eclipse forceplate info')
            fp_info_auto = {k: 'Auto' for k, v in fp_info.items()}
            eclipse.set_eclipse_keys(enffile, fp_info_auto,
                                     update_existing=True)

    # all preprocessing done
    # compute velocity thresholds using all trials
    vel_th = {key: (np.median(x) if x.size > 0 else None) for key, x in
              foot_vel.items()}

    # 2nd pass
    sel_trials = {filepath: trial for filepath, trial in trials.items()
                  if trial['recon_ok']}
    logger.debug('\n2nd pass - processing %d trials\n' % len(sel_trials))

    for filepath, trial in sel_trials.items():
        filename = os.path.split(filepath)[1]
        logger.debug('loading in Nexus: %s' % filename)
        vicon.OpenTrial(filepath, cfg.autoproc.nexus_timeout)
        enf_file = filepath + '.Trial.enf'

        # automark using global velocity thresholds
        try:
            vicon.ClearAllEvents()
            evs = utils.automark_events(vicon, vel_thresholds=vel_th,
                                        mkrdata=trial['mkrdata'],
                                        fp_events=trial['fpev'],
                                        events_range=cfg.autoproc.events_range,
                                        start_on_forceplate=cfg.autoproc.
                                        start_on_forceplate, roi=trial['roi'])
        except GaitDataError:  # cannot automark
            eclipse_str = '%s,%s' % (trial['description'],
                                     cfg.autoproc.enf_descriptions
                                     ['automark_failure'])
            logger.debug('automark failed')
            _save_trial()
            trial['description'] = eclipse_str
            continue  # next trial

        # crop trial around events
        if nexus_ver >= 2.5:
            evs_all = list(itertools.chain.from_iterable(evs.values()))
            if evs_all:
                # when setting roi, do not go beyond trial range
                minfr, maxfr = vicon.GetTrialRange()
                roistart = max(min(evs_all) - cfg.autoproc.crop_margin, minfr)
                roiend = min(max(evs_all) + cfg.autoproc.crop_margin, maxfr)
                # method cannot take numpy.int64
                vicon.SetTrialRegionOfInterest(int(roistart), int(roiend))

        # run model pipeline and save
        eclipse_str = '%s,%s' % (cfg.autoproc.enf_descriptions['ok'],
                                 trial['description'])
        nexus.run_pipelines(vicon, cfg.autoproc.model_pipelines)
        _save_trial()
        trial['description'] = eclipse_str

    # all done; update Eclipse descriptions
    if cfg.autoproc.eclipse_write_key and update_eclipse:
        # try to avoid a possible race condition where Nexus is still
        # holding the .enf file open
        time.sleep(.5)
        for filepath, trial in trials.items():
            enf_file = filepath + '.Trial.enf'
            try:
                eclipse.set_eclipse_keys(enf_file,
                                         {cfg.autoproc.eclipse_write_key:
                                          trial['description']},
                                         update_existing=True)
            except IOError:
                logger.warning('Could not write Eclipse description to %s' %
                               enf_file)
    else:
        logger.debug('not updating Eclipse data')

    # print stats
    logger.debug('Complete')
    logger.debug('Trials opened: %d' % len(trials))
    logger.debug('Trials with recon ok: %d' % len(sel_trials))


def _delete_c3ds(enffiles):
    """ c3d files need to be deleted before processing. Otherwise Nexus will
    load analog data from existing c3d files which are affected by previous
    crop operations, e.g. forceplate data might be clipped """
    logger.debug('deleting previous c3d files')
    c3dfiles = sessionutils._filter_to_c3ds(enffiles)
    for enffile, c3dfile in zip(enffiles, c3dfiles):
        if not op.isfile(c3dfile):
            continue
        edata = eclipse.get_eclipse_keys(enffile, return_empty=True)
        # do not delete static .c3d files (needed for dynamic processing)
        if edata['TYPE'] == 'Static':
            logger.debug('keeping static c3d file %s' % c3dfile)
            continue

        # to prevent data loss, do not delete c3d if original
        # x1d and x2d do not exist
        x1dfile = sessionutils._enf2other(enffile, 'x1d')
        x2dfile = sessionutils._enf2other(enffile, 'x2d')
        if (op.isfile(x1dfile) and op.isfile(x2dfile)):
            logger.debug('deleting existing c3d file %s' % c3dfile)
            os.remove(c3dfile)
        else:
            logger.debug('refusing to delete c3d file %s since original '
                         'data files .(x1d and .x2d) do not exist' % c3dfile)


def autoproc_session(patterns=None, update_eclipse=True):

    sessionpath = nexus.get_sessionpath()
    enffiles = list(sessionutils.get_session_enfs(sessionpath))

    if not enffiles:
        raise GaitDataError('No trials found (no .enf files in session)')

    _delete_c3ds(enffiles)

    if patterns:
        # filter trial names according to patterns
        enffiles = [s for s in enffiles if any([p in s for p in patterns])]
    if enffiles:
        _do_autoproc(enffiles, update_eclipse=update_eclipse)


if __name__ == '__main__':

    logging.basicConfig(level=logging.DEBUG)

    parser = argparse.ArgumentParser()
    parser.add_argument('--include', metavar='p', type=str, nargs='+',
                        help='strings that must appear in trial name')
    parser.add_argument('--no_eclipse', action='store_true',
                        help='disable writing of Eclipse entries')

    args = parser.parse_args()
    autoproc_session(args.include, update_eclipse=not args.no_eclipse)



# -*- coding: utf-8 -*-
"""
@author: Jussi (jnu@iki.fi)
"""

from setuptools import setup, find_packages


setup(name='gaitutils',
      version='0.11.30',
      description='Utilities for processing and plotting gait data',
      author='Jussi Nurminen',
      author_email='jnu@iki.fi',
      license='GPLv3',
      url='https://github.com/jjnurminen/gaitutils',
      packages=find_packages(),
      scripts=['gaitutils/scripts/gaitmenu.py',
               'gaitutils/scripts/tardieu.py',
               'gaitutils/scripts/plotter_gui.py',
               'gaitutils/scripts/nexus_plot.py',
               'gaitutils/scripts/nexus_kin_consistency.py',
               'gaitutils/scripts/nexus_emg_consistency.py',
               'gaitutils/scripts/nexus_musclelen_consistency.py',
               'gaitutils/scripts/nexus_automark_trial.py',
               'gaitutils/scripts/nexus_autoprocess_session.py',
               'gaitutils/scripts/nexus_autoprocess_trial.py',
               'gaitutils/scripts/nexus_copy_trial_videos.py',
               'gaitutils/scripts/nexus_trials_velocity.py',
               'gaitutils/scripts/nexus_make_pdf_report.py',
               'gaitutils/scripts/nexus_make_comparison_report.py',
               'gaitutils/scripts/nexus_kin_average.py',
               'gaitutils/scripts/nexus_time_distance_vars.py'],
      include_package_data=True,
      )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Plot time-distance vars from single trial or as average of tagged trials

@author: Jussi (jnu@iki.fi)
"""

from builtins import zip
from builtins import range
import logging
import argparse
import os.path as op
import matplotlib.pyplot as plt
from matplotlib.backends.backend_pdf import PdfPages
import numpy as np

from gaitutils import (nexus, register_gui_exception_handler, analysis,
                       GaitDataError, sessionutils, cfg)
from gaitutils.plot import time_dist_barchart

logger = logging.getLogger(__name__)


def do_session_average_plot(sessionpath=None, tags=None, show=True,
                            make_pdf=True):
    """Find tagged trials from current session dir and plot average"""

    if sessionpath is None:
        sessionpath = nexus.get_sessionpath()
    tags = tags or cfg.eclipse.tags
    trials = sessionutils.get_c3ds(sessionpath, tags=tags,
                                   trial_type='dynamic')
    if not trials:
        raise GaitDataError('No marked trials found for session %s'
                            % sessionpath)
    fig = _plot_trials([trials])
    session = op.split(sessionpath)[-1]
    fig.suptitle('Time-distance average, session %s' % session)

    if make_pdf:
        pdf_name = op.join(sessionpath, 'time_distance_average.pdf')
        with PdfPages(pdf_name) as pdf:
            pdf.savefig(fig)

    if show:
        plt.show()

    return fig


def do_single_trial_plot(c3dfile, show=True, make_pdf=True):
    """Plot a single trial time-distance."""

    fig = _plot_trials([[c3dfile]])
    c3dpath, c3dfile_ = op.split(c3dfile)
    fig.suptitle('Time-distance variables, %s' % c3dfile_)

    if make_pdf:
        c3dfile_root = op.splitext(c3dfile_)[0]
        pdf_name = op.join(c3dpath, 'time_distance_%s.pdf' % c3dfile_root)
        with PdfPages(pdf_name) as pdf:
            pdf.savefig(fig)

    if show:
        plt.show()

    return fig


def do_multitrial_plot(c3dfiles, show=True, make_pdf=True):
    """Plot multiple trial comparison time-distance.
    PDF goes into Nexus session dir"""

    cond_labels = [op.split(c3d)[-1] for c3d in c3dfiles]
    fig = _plot_trials([[c3d] for c3d in c3dfiles], cond_labels)

    if make_pdf:
        pdf_name = op.join(nexus.get_sessionpath(),
                           '%representative_time_distance.pdf')
        with PdfPages(pdf_name) as pdf:
            pdf.savefig(fig)

    if show:
        plt.show()

    return fig


def do_comparison_plot(sessions, tags, show=True):
    """Time-dist comparison of multiple sessions. Tagged trials from each
    session will be picked."""

    trials = list()
    for session in sessions:
        c3ds = sessionutils.get_c3ds(session, tags=tags, trial_type='dynamic')
        if not c3ds:
            raise ValueError('No tagged trials found in session %s' % session)
        trials.append(c3ds)

    cond_labels = [op.split(session)[-1] for session in sessions]
    fig = _plot_trials(trials, cond_labels)

    if show:
        plt.show()

    return fig


def _plot_trials(trials, cond_labels=None, interactive=True):
    """Make a time-distance variable barchart from given trials (.c3d files).
    trials: list of lists, where inner lists represent conditions
    and list elements represent trials.
    Conditions is a matching list of condition labels.
    If there are multiple trials per condition, they will be averaged.
    plotvars specifies variables to plot (if not all) and their order
    """
    # XXX: hardcoded time-distance variables, to set a certain order
    # set plotvars=None to plot all analysis variables from c3d
    plotvars = ['Walking Speed', 'Cadence', 'Foot Off', 'Opposite Foot Off',
                'Opposite Foot Contact', 'Double Support', 'Single Support',
                'Stride Time', 'Stride Length', 'Step Width', 'Step Length']

    if cond_labels is None:
        cond_labels = ['Condition %d' % k for k in range(len(trials))]
    res_avg_all = dict()
    res_std_all = dict()
    for cond_files, cond_label in zip(trials, cond_labels):
        ans = list()
        for c3dfile in cond_files:
            an = analysis.get_analysis(c3dfile, condition=cond_label)
            ans.append(an)
        if len(ans) > 1:  # do average for this condition
            res_avg = analysis.group_analysis(ans)
            res_std = analysis.group_analysis(ans, fun=np.std)
        else:  # do single-trial plot for this condition
            res_avg = ans[0]
            res_std = dict()
            res_std[cond_label] = None
        res_avg_all.update(res_avg)
        res_std_all.update(res_std)

    return time_dist_barchart(res_avg_all, stddev=res_std_all,
                              stddev_bars=False, plotvars=plotvars,
                              interactive=interactive)


if __name__ == '__main__':

    parser = argparse.ArgumentParser()
    parser.add_argument('--tags', metavar='p', type=str, nargs='+',
                        help='strings that must appear in trial '
                        'description or notes')
    args = parser.parse_args()
    logging.basicConfig(level=logging.DEBUG)
    register_gui_exception_handler()
    do_session_average_plot(tags=args.tags)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Embed gaitplotter figure canvas into PyQt5 GUI
WIP


@author: Jussi (jnu@iki.fi)
"""

from builtins import str
import logging
import sys
import traceback
from pkg_resources import resource_filename
from PyQt5 import QtWidgets, uic, QtCore
from matplotlib.backends.backend_qt5agg import (FigureCanvasQTAgg as
                                                FigureCanvas)
from matplotlib.backends.backend_qt5agg import (NavigationToolbar2QT as
                                                NavigationToolbar)

from .. import (Plotter, Trial, nexus, layouts, cfg, GaitDataError,
                stats)
from .qt_dialogs import qt_message_dialog, OptionsDialog

logger = logging.getLogger(__name__)


def _trial_namestr(trial):
    return u'%s  %s  %s' % (trial.trialname, trial.eclipse_data['DESCRIPTION'],
                            trial.eclipse_data['NOTES'])


class AveragerDialog(QtWidgets.QDialog):

    def __init__(self, parent=None):
        super(self.__class__, self).__init__()
        uifile = resource_filename('gaitutils', 'gui/averager.ui')
        uic.loadUi(uifile, self)

        self.btnAddNexusTrial.clicked.connect(self._open_nexus_trial)
        self.btnAddC3DTrial.clicked.connect(self.load_dialog)
        self.btnDoAverage.clicked.connect(self.accept)
        self.btnCancel.clicked.connect(self.reject)

    def _open_nexus_trial(self):
        try:
            vicon = nexus.viconnexus()
            self._open_trial(vicon)
        except GaitDataError:
            qt_message_dialog('Vicon Nexus is not running')

    def _open_trial(self, source):
        tr = Trial(source)
        trials = (item.userdata for item in self.listTrials.items)
        # check if trial already loaded (based on name)
        # TODO: might use smarter detection
        if tr.trialname in [trial.trialname for trial in trials]:
            return
        self.listTrials.add_item(_trial_namestr(tr), data=tr)

    def accept(self):
        """ Do the averaging """
        trials = [item.userdata for item in self.listTrials.items]
        fp_cycles_only = self.xpFpCyclesOnly.checkState()
        self.avg = stats.AvgTrial(trials, fp_cycles_only=fp_cycles_only)
        qt_message_dialog('Averaging OK')
        self.done(QtWidgets.QDialog.Accepted)  # or call superclass accept

    def load_dialog(self):
        """ Bring up load dialog and load selected c3d file. """
        fnames = QtWidgets.QFileDialog.getOpenFileNames(self, 'Open C3D files',
                                                        '*.c3d')[0]
        if fnames:
            for fname in fnames:
                self._open_trial(fname)


class PlotterWindow(QtWidgets.QMainWindow):

    def __init__(self, parent=None):
        super(PlotterWindow, self).__init__(parent)

        uifile = resource_filename('gaitutils', 'gui/plotter_gui.ui')
        uic.loadUi(uifile, self)

        self.pl = Plotter(interactive=False)
        self.canvas = FigureCanvas(self.pl.fig)
        self.canvas.setSizePolicy(QtWidgets.QSizePolicy.Expanding,
                                  QtWidgets.QSizePolicy.Expanding)

        # these are needed for mpl callbacks to work (?)
        self.canvas.setFocusPolicy(QtCore.Qt.ClickFocus)
        self.canvas.setFocus()

        # self.setStyleSheet("background-color: white;");
        # add canvas into last column, span all rows
        self.mainGridLayout.addWidget(self.canvas, 0,
                                      self.mainGridLayout.columnCount(),
                                      self.mainGridLayout.rowCount(), 1)

        # this is the Navigation widget
        # it takes the Canvas widget and a parent
        #self.toolbar = NavigationToolbar(self.canvas, self)
        # can use this to enable selection of multiple items:
        # self.listTrials.setSelectionMode(QtWidgets.QAbstractItemView.MultiSelection)
        self.listTrials.itemPressed.connect(self._trial_selected)
        # set up widgets
        # buttons
        self.btnAddNexusTrial.clicked.connect(self._open_nexus_trial)
        self.btnPlot.clicked.connect(self._plot_trials)
        self.btnSelectAllCycles.clicked.connect(self.listTrialCycles.check_all)
        self.btnSelectForceplateCycles.clicked.connect(self.
                                                       _select_forceplate_cycles)
        self.btnSelect1stCycles.clicked.connect(self._select_1st_cycles)
        self.btnPickCycles.clicked.connect(self._pick_cycles)
        #self.btnAverageSelected.clicked.connect(self._average_selected)
        self.btnAddC3DTrial.clicked.connect(self.load_dialog)
        self.btnSavePDF.clicked.connect(self._write_pdf)
        self.btnClearTrials.clicked.connect(self._clear_trials)
        self.btnClearCyclesToPlot.clicked.connect(self.listCyclesToPlot.clear)
        self.btnOptions.clicked.connect(self._options_dialog)
        self.btnOpenNormalData.clicked.connect(self._load_normaldata)
        # menu actions
        self.actionQuit.triggered.connect(self.close)
        self.actionAverage.triggered.connect(self._averager_dialog)

        # add predefined plot layouts to combobox
        self.cbLayout.addItems(sorted(cfg.options('layouts')))

        # add normal data files
        self.cbNormalData.addItems(sorted(cfg.general.normaldata_files))

        self.canvas.mpl_connect('button_press_event', self._onclick)
        self._set_status('Ready')

    def resizeEvent(self, event):
        pass
        # adjust layout on resize
        # this does not work super well - slow and can crash
        # self.pl.tight_layout()

    def _onclick(self, event):
        logger.debug('click on %s' % event.inaxes)

    def _options_dialog(self):
        """ Show the autoprocessing options dialog """
        dlg = OptionsDialog(default_tab=1)
        dlg.exec_()

    def _averager_dialog(self):
        """ Show the autoprocessing options dialog """
        dlg = AveragerDialog()
        if dlg.exec_():
            self.listTrials.add_item('Average trial', dlg.avg)
            self._update_trial_cycles_list(dlg.avg)

    def _load_normaldata(self):
        fname = QtWidgets.QFileDialog.getOpenFileName(self,
                                                      'Open normal data', '',
                                                      'XLSX files (*.xlsx);;'
                                                      'GCD files (*.gcd);; ')
        fname = fname[0]
        if fname:
            idx = self.cbNormalData.findText(fname)
            # add file into box if it's not yet there
            if idx == -1:
                self.cbNormalData.addItem(fname)
            # select it in any case
            idx = self.cbNormalData.findText(fname)
            self.cbNormalData.setCurrentIndex(idx)

    def rm_trial(self):
        if self.listTrials.currentItem():
            self.listTrials.rm_current_item()
            self.listTrialCycles.clear()

    def _set_status(self, msg):
        self.statusbar.showMessage(msg)

    def _trial_selected(self, item):
        """ User clicked on trial """
        trial = item.userdata
        self._update_trial_cycles_list(trial)

    def _update_trial_cycles_list(self, trial):
        """ Show cycles of given trial in the trial cycle list """
        cycles = trial.cycles
        self.listTrialCycles.clear()
        for cycle in cycles:
            self.listTrialCycles.add_item(cycle.name, data=cycle,
                                          checkable=True)

    def _clear_trials(self):
        self.listTrials.clear()
        self.listTrialCycles.clear()

    def _select_forceplate_cycles(self):
        """ Select all forceplate cycles in the trial cycles list """
        self.listTrialCycles.check_none()
        for item in self.listTrialCycles.items:
            if item.userdata.on_forceplate:
                item.checkstate = True

    def _select_1st_cycles(self):
        """ Select 1st L/R cycles in the trial cycles list """
        self.listTrialCycles.check_none()
        for item in self.listTrialCycles.items:
            if item.userdata.index == 1:
                item.checkstate = True

    def _pick_cycles(self):
        """ Add selected cycles to overall list """
        # only add cycles that were not already added
        present_cycles = (item.userdata for item in
                          self.listCyclesToPlot.items)
        new_cycles = (item.userdata for item in
                      self.listTrialCycles.checked_items)
        for cycle in set(new_cycles) - set(present_cycles):
            name = '%s: %s' % (cycle.trial.trialname, cycle.name)
            self.listCyclesToPlot.add_item(name, data=cycle)

    def _open_nexus_trial(self):
        try:
            vicon = nexus.viconnexus()
            self._open_trial(vicon)
        except GaitDataError:
            qt_message_dialog('Vicon Nexus is not running')

    def _open_trial(self, source):
        tr = Trial(source)
        trials = (item.userdata for item in self.listTrials.items)
        # check if trial already loaded (based on name)
        # TODO: might use smarter detection
        if tr.trialname in [trial.trialname for trial in trials]:
            return
        self.listTrials.add_item(_trial_namestr(tr), data=tr)
        self._update_trial_cycles_list(tr)
        self._set_status('Loaded trial %s' % tr.trialname)

    def _plot_trials(self):
        """ Plot user-picked trials """
        # Holds cycles to plot, keyed by trial. This will make for less calls
        # to plot_trial() as we only need to call once per trial.
        plot_cycles = dict()
        self.pl.fig.clear()

        cycles = [item.userdata for item in self.listCyclesToPlot.items]
        if not cycles:
            qt_message_dialog('No cycles selected for plotting')
            return

        # gather the cycles and key by trial
        for cycle in cycles:
            if cycle.trial not in plot_cycles:
                plot_cycles[cycle.trial] = []
            plot_cycles[cycle.trial].append(cycle)

        # set options and create the plot
        lout = cfg.layouts.__getattr__(self.cbLayout.currentText())
        # remove legends - they do not make sense here
        lout = layouts.filter_layout(lout, 'legend', None)
        self.pl.layout = lout
        match_pig_kinetics = self.xbKineticsFpOnly.checkState()

        if self.xbNormalData.checkState():
            plot_model_normaldata = True
            try:
                fn = self.cbNormalData.currentText()
                self.pl.add_normaldata(fn)
            except (ValueError, GaitDataError):
                qt_message_dialog('Error reading normal data from %s' % fn)
                return
        else:
            plot_model_normaldata = False

        for trial, cycs in plot_cycles.items():
            try:
                self.pl.plot_trial(trial=trial, model_cycles=cycs,
                                   emg_cycles=cycs,
                                   match_pig_kinetics=match_pig_kinetics,
                                   maintitle='', superpose=True,
                                   model_stddev=trial.stddev_data,
                                   plot_model_normaldata=plot_model_normaldata)
            except GaitDataError as e:
                qt_message_dialog('Error: %s' % str(e))
                self.pl.fig.clear()  # fig may have been partially drawn
                return

        self.canvas.draw()
        self.btnSavePDF.setEnabled(True)  # can create pdf now
        # this is a lazy default for sessionpath (pick path from one trial)
        # it is used as the default PDF output dir.
        self._sessionpath = trial.sessionpath

    def load_dialog(self):
        """ Bring up load dialog and load selected c3d file. """
        fnames = QtWidgets.QFileDialog.getOpenFileNames(self, 'Open C3D files',
                                                        '*.c3d')[0]
        if fnames:
            for fname in fnames:
                self._open_trial(fname)

    def _write_pdf(self):
        """ Bring up save dialog and save data. """
        fname = QtWidgets.QFileDialog.getSaveFileName(self,
                                                      'Save PDF',
                                                      self._sessionpath,
                                                      '*.pdf')[0]
        if fname:
            # TODO: need to figure out logic for the title
            self.pl.set_title('Just a test')
            old_size = self.pl.fig.get_size_inches()
            self.pl.fig.set_size_inches([8.27, 11.69])
            try:
                self.canvas.print_figure(fname)
            except IOError:
                qt_message_dialog('Error writing PDF file, check that file '
                                  'is not open')
            # reset title for onscreen and redraw canvas
            self.pl.set_title('')
            self.pl.fig.set_size_inches(old_size)  # needed?
            self.canvas.draw()


def main():

    logging.basicConfig(level=logging.DEBUG)
    # uiparser logger makes too much noise
    logging.getLogger('PyQt5.uic').setLevel(logging.WARNING)

    app = QtWidgets.QApplication(sys.argv)
    win = PlotterWindow()

    def my_excepthook(type_, value, tback):
        """ Custom exception handler for fatal (unhandled) exceptions:
        report to user via GUI and terminate. """
        tb_full = u''.join(traceback.format_exception(type_, value, tback))
        qt_message_dialog('Unhandled exception: %s' % tb_full)
        # dump traceback to file
        # try:
        #    with io.open(Config.traceback_file, 'w', encoding='utf-8') as f:
        #        f.write(tb_full)
        # here is a danger of infinitely looping the exception hook,
        # so try to catch any exceptions...
        # except Exception:
        #    print('Cannot dump traceback!')
        sys.__excepthook__(type_, value, tback)
        app.quit()

    sys.excepthook = my_excepthook

    win.show()
    sys.exit(app.exec_())
